from django.utils.translation import ugettext_lazy as _
from rest_framework import decorators, response, status, serializers as rf_serializers

from waldur_core.core import executors as core_executors, validators as core_validators
from waldur_core.structure import views as structure_views

from . import models, serializers, log, filters, executors


class DigitalOceanServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.DigitalOceanService.objects.all()
    serializer_class = serializers.ServiceSerializer
    import_serializer_class = serializers.DropletImportSerializer


class DigitalOceanServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.DigitalOceanServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer


class ImageViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Image.objects.all()
    serializer_class = serializers.ImageSerializer
    filterset_class = filters.ImageFilter
    lookup_field = 'uuid'


class RegionViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Region.objects.all()
    serializer_class = serializers.RegionSerializer
    filterset_class = filters.RegionFilter
    lookup_field = 'uuid'

    def get_queryset(self):
        return models.Region.objects.order_by('name')


class SizeViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Size.objects.all()
    serializer_class = serializers.SizeSerializer
    filterset_class = filters.SizeFilter
    lookup_field = 'uuid'


class DropletViewSet(structure_views.ResourceViewSet):
    queryset = models.Droplet.objects.all()
    serializer_class = serializers.DropletSerializer
    filterset_class = filters.DropletFilter
    create_executor = executors.DropletCreateExecutor
    update_executor = core_executors.EmptyExecutor
    delete_executor = executors.DropletDeleteExecutor
    destroy_validators = [core_validators.StateValidator(models.Droplet.States.OK, models.Droplet.States.ERRED)]

    def perform_create(self, serializer):
        region = serializer.validated_data['region']
        image = serializer.validated_data['image']
        size = serializer.validated_data['size']
        ssh_key = serializer.validated_data.get('ssh_public_key')

        droplet = serializer.save(
            cores=size.cores,
            ram=size.ram,
            disk=size.disk,
            transfer=size.transfer)

        # XXX: We do not operate with backend_id`s in views.
        #      View should pass objects to executor.
        self.create_executor.execute(
            droplet,
            is_async=self.async_executor,
            backend_region_id=region.backend_id,
            backend_image_id=image.backend_id,
            backend_size_id=size.backend_id,
            ssh_key_uuid=ssh_key.uuid.hex if ssh_key else None)

    @decorators.action(detail=True, methods=['post'])
    def start(self, request, uuid=None):
        instance = self.get_object()
        executors.DropletStartExecutor().execute(instance)
        return response.Response({'status': _('start was scheduled')}, status=status.HTTP_202_ACCEPTED)

    start_validators = [core_validators.StateValidator(models.Droplet.States.OK),
                        core_validators.RuntimeStateValidator(models.Droplet.RuntimeStates.OFFLINE)]
    start_serializer_class = rf_serializers.Serializer

    @decorators.action(detail=True, methods=['post'])
    def stop(self, request, uuid=None):
        instance = self.get_object()
        executors.DropletStopExecutor().execute(instance)
        return response.Response({'status': _('stop was scheduled')}, status=status.HTTP_202_ACCEPTED)

    stop_validators = [core_validators.StateValidator(models.Droplet.States.OK),
                       core_validators.RuntimeStateValidator(models.Droplet.RuntimeStates.ONLINE)]
    stop_serializer_class = rf_serializers.Serializer

    @decorators.action(detail=True, methods=['post'])
    def restart(self, request, uuid=None):
        instance = self.get_object()
        executors.DropletRestartExecutor().execute(instance)
        return response.Response({'status': _('restart was scheduled')}, status=status.HTTP_202_ACCEPTED)

    restart_validators = [core_validators.StateValidator(models.Droplet.States.OK),
                          core_validators.RuntimeStateValidator(models.Droplet.RuntimeStates.ONLINE)]
    restart_serializer_class = rf_serializers.Serializer

    @decorators.action(detail=True, methods=['post'])
    def resize(self, request, uuid=None):
        """
        To resize droplet, submit a **POST** request to the instance URL, specifying URI of a target size.

        Pass {'disk': true} along with target size in order to perform permanent resizing,
        which allows you to resize your disk space as well as CPU and RAM.
        After increasing the disk size, you will not be able to decrease it.

        Pass {'disk': false} along with target size in order to perform flexible resizing,
        which only upgrades your CPU and RAM. This option is reversible.

        Note that instance must be OFFLINE. Example of a valid request:

        .. code-block:: http

            POST /api/digitalocean-droplets/6c9b01c251c24174a6691a1f894fae31/resize/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "size": "http://example.com/api/digitalocean-sizes/1ee385bc043249498cfeb8c7e3e079f0/"
            }
        """
        droplet = self.get_object()
        serializer = self.get_serializer(droplet, data=request.data)
        serializer.is_valid(raise_exception=True)

        size = serializer.validated_data['size']
        disk = serializer.validated_data['disk']

        executors.DropletResizeExecutor.execute(
            droplet,
            disk=disk,
            size=size,
            updated_fields=None,
            is_async=self.async_executor)

        message = _('Droplet {droplet_name} has been scheduled to %s resize.') % \
            (disk and _('permanent') or _('flexible'))
        log.event_logger.droplet_resize.info(
            message,
            event_type='droplet_resize_scheduled',
            event_context={'droplet': droplet, 'size': size}
        )

        cores_increment = size.cores - droplet.cores
        ram_increment = size.ram - droplet.ram
        disk_increment = None

        droplet.cores = size.cores
        droplet.ram = size.ram

        if disk:
            disk_increment = size.disk - droplet.disk
            droplet.disk = size.disk

        droplet.save()

        spl = droplet.service_project_link

        if disk_increment:
            spl.add_quota_usage(spl.Quotas.storage, disk_increment, validate=True)
        spl.add_quota_usage(spl.Quotas.ram, ram_increment, validate=True)
        spl.add_quota_usage(spl.Quotas.vcpu, cores_increment, validate=True)

        return response.Response({'detail': _('resizing was scheduled')}, status=status.HTTP_202_ACCEPTED)

    resize_validators = [core_validators.StateValidator(models.Droplet.States.OK)]
    resize_serializer_class = serializers.DropletResizeSerializer



from . import views


def register_in(router):
    router.register(r'rijkscloud', views.ServiceViewSet, basename='rijkscloud')
    router.register(r'rijkscloud-service-project-link', views.ServiceProjectLinkViewSet,
                    basename='rijkscloud-spl')
    router.register(r'rijkscloud-flavors', views.FlavorViewSet, basename='rijkscloud-flavor')
    router.register(r'rijkscloud-volumes', views.VolumeViewSet, basename='rijkscloud-volume')
    router.register(r'rijkscloud-instances', views.InstanceViewSet, basename='rijkscloud-instance')
    router.register(r'rijkscloud-networks', views.NetworkViewSet, basename='rijkscloud-network')
    router.register(r'rijkscloud-subnets', views.SubNetViewSet, basename='rijkscloud-subnet')
    router.register(r'rijkscloud-internal-ips', views.InternalIPViewSet, basename='rijkscloud-internal-ip')
    router.register(r'rijkscloud-floating-ips', views.FloatingIPViewSet, basename='rijkscloud-fip')



import datetime
from functools import reduce
import itertools

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.core.validators import MaxLengthValidator
from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models, transaction
from django.db.models import Q, signals
from django.utils import timezone
from django.utils.lru_cache import lru_cache
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker
from model_utils.fields import AutoCreatedField
from model_utils.models import TimeStampedModel, SoftDeletableModel
from model_utils.managers import SoftDeletableManagerMixin
import pyvat

from waldur_core.core import fields as core_fields
from waldur_core.core import models as core_models
from waldur_core.core import utils as core_utils
from waldur_core.core.fields import JSONField
from waldur_core.core.models import AbstractFieldTracker
from waldur_core.core.shims import TaggableManager
from waldur_core.core.validators import validate_name, validate_cidr_list
from waldur_core.media.validators import CertificateValidator
from waldur_core.logging.loggers import LoggableMixin
from waldur_core.monitoring.models import MonitoringModelMixin
from waldur_core.quotas import models as quotas_models, fields as quotas_fields
from waldur_core.structure import SupportedServices
from waldur_core.media.models import ImageModelMixin
from waldur_core.structure.managers import StructureManager, filter_queryset_for_user, \
    ServiceSettingsManager, PrivateServiceSettingsManager, SharedServiceSettingsManager
from waldur_core.structure.signals import structure_role_granted, structure_role_revoked
from waldur_core.structure.utils import sort_dependencies
from waldur_geo_ip.utils import get_coordinates_by_ip
from waldur_geo_ip.mixins import IPCoordinatesMixin, CoordinatesMixin


def validate_service_type(service_type):
    from django.core.exceptions import ValidationError
    if not SupportedServices.has_service_type(service_type):
        raise ValidationError(_('Invalid service type.'))


class StructureModel(models.Model):
    """ Generic structure model.
        Provides transparent interaction with base entities and relations like customer.
    """

    objects = StructureManager()

    class Meta:
        abstract = True

    def __getattr__(self, name):
        # add additional properties to the object according to defined Permissions class
        fields = ('customer', 'project')
        if name in fields:
            try:
                path = getattr(self.Permissions, name + '_path')
            except AttributeError:
                pass
            else:
                if not path == 'self' and '__' in path:
                    return reduce(getattr, path.split('__'), self)

        raise AttributeError(
            "'%s' object has no attribute '%s'" % (self._meta.object_name, name))


class StructureLoggableMixin(LoggableMixin):

    @classmethod
    def get_permitted_objects(cls, user):
        return filter_queryset_for_user(cls.objects.all(), user)


class TagMixin(models.Model):
    """
    Add tags field and manage cache for tags.
    """
    class Meta:
        abstract = True

    tags = TaggableManager(blank=True, related_name='+')

    def get_tags(self):
        key = self._get_tag_cache_key()
        tags = cache.get(key)
        if tags is None:
            tags = list(self.tags.all().values_list('name', flat=True))
            cache.set(key, tags)
        return tags

    def clean_tag_cache(self):
        key = self._get_tag_cache_key()
        cache.delete(key)

    def _get_tag_cache_key(self):
        return 'tags:%s' % core_utils.serialize_instance(self)


class VATException(Exception):
    pass


class VATMixin(models.Model):
    """
    Add country, VAT number fields and check results from EU VAT Information Exchange System.
    Allows to compute VAT charge rate.
    """
    class Meta:
        abstract = True

    vat_code = models.CharField(max_length=20, blank=True, help_text=_('VAT number'))
    vat_name = models.CharField(max_length=255, blank=True,
                                help_text=_('Optional business name retrieved for the VAT number.'))
    vat_address = models.CharField(max_length=255, blank=True,
                                   help_text=_('Optional business address retrieved for the VAT number.'))

    is_company = models.BooleanField(default=False, help_text=_('Is company or private person'))
    country = core_fields.CountryField(blank=True)

    def get_vat_rate(self):
        charge = self.get_vat_charge()
        if charge.action == pyvat.VatChargeAction.charge:
            return charge.rate

        # Return None, if reverse_charge or no_charge action is applied

    def get_vat_charge(self):
        if not self.country:
            raise VATException(_('Unable to get VAT charge because buyer country code is not specified.'))

        seller_country = settings.WALDUR_CORE.get('SELLER_COUNTRY_CODE')
        if not seller_country:
            raise VATException(_('Unable to get VAT charge because seller country code is not specified.'))

        return pyvat.get_sale_vat_charge(
            datetime.date.today(),
            pyvat.ItemType.generic_electronic_service,
            pyvat.Party(self.country, self.is_company and self.vat_code),
            pyvat.Party(seller_country, True)
        )


class BasePermission(models.Model):
    class Meta:
        abstract = True

    user = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, db_index=True)
    created_by = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, null=True, blank=True, related_name='+')
    created = AutoCreatedField()
    expiration_time = models.DateTimeField(null=True, blank=True)
    is_active = models.NullBooleanField(default=True, db_index=True)

    @classmethod
    def get_url_name(cls):
        raise NotImplementedError

    @classmethod
    def get_expired(cls):
        return cls.objects.filter(expiration_time__lt=timezone.now(), is_active=True)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    def revoke(self):
        raise NotImplementedError


class PermissionMixin:
    """
    Base permission management mixin for customer and project.
    It is expected that reverse `permissions` relation is created for this model.
    Provides method to grant, revoke and check object permissions.
    """

    def has_user(self, user, role=None, timestamp=False):
        """
        Checks whether user has role in entity.
        `timestamp` can have following values:
            - False - check whether user has role in entity at the moment.
            - None - check whether user has permanent role in entity.
            - Datetime object - check whether user will have role in entity at specific timestamp.
        """
        permissions = self.permissions.filter(user=user, is_active=True)

        if role is not None:
            permissions = permissions.filter(role=role)

        if timestamp is None:
            permissions = permissions.filter(expiration_time=None)
        elif timestamp:
            permissions = permissions.filter(Q(expiration_time=None) | Q(expiration_time__gte=timestamp))

        return permissions.exists()

    @transaction.atomic()
    def add_user(self, user, role, created_by=None, expiration_time=None):
        permission = self.permissions.filter(user=user, role=role, is_active=True).first()
        if permission:
            return permission, False

        permission = self.permissions.create(
            user=user,
            role=role,
            is_active=True,
            created_by=created_by,
            expiration_time=expiration_time,
        )

        structure_role_granted.send(
            sender=self.__class__,
            structure=self,
            user=user,
            role=role,
            created_by=created_by,
        )

        return permission, True

    @transaction.atomic()
    def remove_user(self, user, role=None, removed_by=None):
        permissions = self.permissions.all().filter(user=user, is_active=True)

        if role is not None:
            permissions = permissions.filter(role=role)

        affected_permissions = list(permissions)
        permissions.update(is_active=None, expiration_time=timezone.now())

        for permission in affected_permissions:
            self.log_role_revoked(permission, removed_by)

    @transaction.atomic()
    def remove_all_users(self):
        for permission in self.permissions.all().iterator():
            permission.delete()
            self.log_role_revoked(permission)

    def log_role_revoked(self, permission, removed_by=None):
        structure_role_revoked.send(
            sender=self.__class__,
            structure=self,
            user=permission.user,
            role=permission.role,
            removed_by=removed_by,
        )


class CustomerRole(models.CharField):
    OWNER = 'owner'
    SUPPORT = 'support'

    CHOICES = (
        (OWNER, 'Owner'),
        (SUPPORT, 'Support'),
    )

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 30
        kwargs['choices'] = self.CHOICES
        super(CustomerRole, self).__init__(*args, **kwargs)


class CustomerPermission(BasePermission):
    class Meta:
        unique_together = ('customer', 'role', 'user', 'is_active')

    class Permissions:
        customer_path = 'customer'

    customer = models.ForeignKey(on_delete=models.CASCADE, to='structure.Customer', verbose_name=_('organization'), related_name='permissions')
    role = CustomerRole(db_index=True)
    tracker = FieldTracker(fields=['expiration_time'])

    @classmethod
    def get_url_name(cls):
        return 'customer_permission'

    def revoke(self):
        self.customer.remove_user(self.user, self.role)

    def __str__(self):
        return '%s | %s' % (self.customer.name, self.get_role_display())


def get_next_agreement_number():
    initial_number = settings.WALDUR_CORE['INITIAL_CUSTOMER_AGREEMENT_NUMBER']
    last_number = Customer.objects.aggregate(models.Max('agreement_number')).get('agreement_number__max')
    return (last_number or initial_number) + 1


class DivisionType(core_models.UuidMixin,
                   core_models.NameMixin,
                   models.Model):

    class Meta:
        verbose_name = _('division type')
        ordering = ('name',)

    def __str__(self):
        return self.name


class Division(core_models.UuidMixin,
               core_models.NameMixin,
               models.Model):

    type = models.ForeignKey(on_delete=models.CASCADE, to='DivisionType')
    parent = models.ForeignKey(on_delete=models.CASCADE, to='Division', null=True, blank=True)

    class Meta:
        verbose_name = _('division')
        ordering = ('name',)

    @classmethod
    def get_url_name(cls):
        return 'division'

    def __str__(self):
        full_path = [self.name]
        d = self.parent

        while d is not None:
            full_path.append(d.name)
            d = d.parent

        return ' -> '.join(full_path[::-1])


class Customer(core_models.UuidMixin,
               core_models.NameMixin,
               core_models.DescendantMixin,
               quotas_models.ExtendableQuotaModelMixin,
               PermissionMixin,
               VATMixin,
               StructureLoggableMixin,
               ImageModelMixin,
               TimeStampedModel,
               CoordinatesMixin,
               StructureModel):
    class Permissions:
        customer_path = 'self'
        project_path = 'projects'

    native_name = models.CharField(max_length=160, default='', blank=True)
    abbreviation = models.CharField(max_length=12, blank=True)
    contact_details = models.TextField(blank=True, validators=[MaxLengthValidator(500)])
    agreement_number = models.PositiveIntegerField(null=True, blank=True, unique=True)
    email = models.EmailField(_('email address'), max_length=75, blank=True)
    phone_number = models.CharField(_('phone number'), max_length=255, blank=True)
    access_subnets = models.TextField(validators=[validate_cidr_list], blank=True, default='',
                                      help_text=_('Enter a comma separated list of IPv4 or IPv6 '
                                                  'CIDR addresses from where connection to self-service is allowed.'))
    backend_id = models.CharField(max_length=255, blank=True, help_text=_(
        'Organization identifier in another application.'
    ))
    registration_code = models.CharField(max_length=160, default='', blank=True)
    homepage = models.URLField(max_length=255, blank=True)
    domain = models.CharField(max_length=255, blank=True)

    type = models.CharField(blank=True, max_length=150)
    address = models.CharField(blank=True, max_length=300)
    postal = models.CharField(blank=True, max_length=20)
    bank_name = models.CharField(blank=True, max_length=150)
    bank_account = models.CharField(blank=True, max_length=50)
    accounting_start_date = models.DateTimeField(_('Start date of accounting'), default=timezone.now)
    default_tax_percent = models.DecimalField(default=0, max_digits=4, decimal_places=2,
                                              validators=[MinValueValidator(0), MaxValueValidator(100)])
    blocked = models.BooleanField(default=False)
    division = models.ForeignKey('Division', null=True, blank=True, on_delete=models.SET_NULL)
    tracker = FieldTracker()

    class Meta:
        verbose_name = _('organization')
        ordering = ('name',)

    GLOBAL_COUNT_QUOTA_NAME = 'nc_global_customer_count'

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        enable_fields_caching = False
        nc_project_count = quotas_fields.CounterQuotaField(
            target_models=lambda: [Project],
            path_to_scope='customer',
        )
        nc_service_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Service.get_all_models(),
            path_to_scope='customer',
        )
        nc_service_project_link_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ServiceProjectLink.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_user_count = quotas_fields.QuotaField()
        nc_resource_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ResourceMixin.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_app_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ApplicationMixin.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_vm_count = quotas_fields.CounterQuotaField(
            target_models=lambda: VirtualMachine.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_private_cloud_count = quotas_fields.CounterQuotaField(
            target_models=lambda: PrivateCloud.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_storage_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Storage.get_all_models(),
            path_to_scope='project.customer',
        )
        nc_volume_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Volume.get_all_models(),
            path_to_scope='customer',
            target_field='size',
        )
        nc_snapshot_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Snapshot.get_all_models(),
            path_to_scope='customer',
            target_field='size',
        )

    def get_log_fields(self):
        return ('uuid', 'name', 'abbreviation', 'contact_details')

    def get_owners(self):
        return get_user_model().objects.filter(
            customerpermission__customer=self,
            customerpermission__is_active=True,
            customerpermission__role=CustomerRole.OWNER,
        )

    def get_support_users(self):
        return get_user_model().objects.filter(
            customerpermission__customer=self,
            customerpermission__is_active=True,
            customerpermission__role=CustomerRole.SUPPORT,
        )

    def get_users(self):
        """ Return all connected to customer users """
        return get_user_model().objects.filter(
            Q(customerpermission__customer=self,
              customerpermission__is_active=True) |
            Q(projectpermission__project__customer=self,
              projectpermission__is_active=True)
        ).distinct().order_by('username')

    def can_user_update_quotas(self, user):
        return user.is_staff

    def can_manage_role(self, user, role=None, timestamp=False):
        """
        Checks whether user can grant/update/revoke customer permissions.
        `timestamp` can have following values:
            - False - check whether user can manage permissions at the moment.
            - None - check whether user can permanently manage permissions.
            - Datetime object - check whether user will be able to manage permissions at specific timestamp.
        """
        return user.is_staff or (
            self.has_user(user, CustomerRole.OWNER, timestamp) and
            settings.WALDUR_CORE['OWNERS_CAN_MANAGE_OWNERS']
        )

    def get_children(self):
        return itertools.chain.from_iterable(
            m.objects.filter(customer=self) for m in [Project] + Service.get_all_models())

    def is_billable(self):
        return timezone.now() >= self.accounting_start_date

    @classmethod
    def get_permitted_objects(cls, user):
        if user.is_staff or user.is_support:
            return cls.objects.all()
        else:
            return cls.objects.filter(
                permissions__user=user,
                permissions__role=CustomerRole.OWNER,
                permissions__is_active=True
            )

    def get_display_name(self):
        if self.abbreviation:
            return self.abbreviation
        if self.domain:
            return '{name} ({domain})'.format(name=self.name, domain=self.domain)
        return self.name

    def delete(self, *args, **kwargs):
        """Delete customers' projects if they all mark as 'removed'."""
        if not self.projects.count():
            for project in Project.structure_objects.filter(customer=self):
                project.delete(soft=False)

        return super(Customer, self).delete(*args, **kwargs)

    def __str__(self):
        if self.abbreviation:
            return '%(name)s (%(abbreviation)s)' % {
                'name': self.name,
                'abbreviation': self.abbreviation
            }
        else:
            return self.name


class ProjectRole(models.CharField):
    ADMINISTRATOR = 'admin'
    MANAGER = 'manager'
    SUPPORT = 'support'

    CHOICES = (
        (ADMINISTRATOR, 'Administrator'),
        (MANAGER, 'Manager'),
        (SUPPORT, 'Support'),
    )

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 30
        kwargs['choices'] = self.CHOICES
        super(ProjectRole, self).__init__(*args, **kwargs)


class ProjectPermission(core_models.UuidMixin, BasePermission):
    class Meta:
        unique_together = ('project', 'role', 'user', 'is_active')

    class Permissions:
        customer_path = 'project__customer'
        project_path = 'project'

    project = models.ForeignKey(on_delete=models.CASCADE, to='structure.Project', related_name='permissions')
    role = ProjectRole(db_index=True)
    tracker = FieldTracker(fields=['expiration_time'])

    @classmethod
    def get_url_name(cls):
        return 'project_permission'

    def revoke(self):
        self.project.remove_user(self.user, self.role)

    def __str__(self):
        return '%s | %s' % (self.project.name, self.get_role_display())


class ProjectType(core_models.DescribableMixin, core_models.UuidMixin, core_models.NameMixin):
    class Meta:
        verbose_name = _('Project type')
        verbose_name_plural = _('Project types')
        ordering = ['name']

    @classmethod
    def get_url_name(cls):
        return 'project_type'

    def __str__(self):
        return self.name


class SoftDeletableManager(SoftDeletableManagerMixin, StructureManager):
    pass


class Project(core_models.DescribableMixin,
              core_models.UuidMixin,
              core_models.NameMixin,
              core_models.DescendantMixin,
              quotas_models.ExtendableQuotaModelMixin,
              PermissionMixin,
              StructureLoggableMixin,
              TimeStampedModel,
              StructureModel,
              SoftDeletableModel):
    class Permissions:
        customer_path = 'customer'
        project_path = 'self'

    GLOBAL_COUNT_QUOTA_NAME = 'nc_global_project_count'

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        enable_fields_caching = False
        nc_resource_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ResourceMixin.get_all_models(),
            path_to_scope='project',
        )
        nc_app_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ApplicationMixin.get_all_models(),
            path_to_scope='project',
        )
        nc_vm_count = quotas_fields.CounterQuotaField(
            target_models=lambda: VirtualMachine.get_all_models(),
            path_to_scope='project',
        )
        nc_private_cloud_count = quotas_fields.CounterQuotaField(
            target_models=lambda: PrivateCloud.get_all_models(),
            path_to_scope='project',
        )
        nc_storage_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Storage.get_all_models(),
            path_to_scope='project',
        )
        nc_volume_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Volume.get_all_models(),
            path_to_scope='project',
        )
        nc_snapshot_count = quotas_fields.CounterQuotaField(
            target_models=lambda: Snapshot.get_all_models(),
            path_to_scope='project',
        )
        nc_service_project_link_count = quotas_fields.CounterQuotaField(
            target_models=lambda: ServiceProjectLink.get_all_models(),
            path_to_scope='project',
        )
        nc_volume_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Volume.get_all_models(),
            path_to_scope='project',
            target_field='size',
        )
        nc_snapshot_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Snapshot.get_all_models(),
            path_to_scope='project',
            target_field='size',
        )

    certifications = models.ManyToManyField(to='ServiceCertification', related_name='projects', blank=True)
    customer = models.ForeignKey(
        Customer, verbose_name=_('organization'), related_name='projects', on_delete=models.PROTECT)
    tracker = FieldTracker()
    type = models.ForeignKey(
        ProjectType, verbose_name=_('project type'), blank=True, null=True, on_delete=models.PROTECT)

    objects = SoftDeletableManager()
    structure_objects = StructureManager()

    @property
    def full_name(self):
        return self.name

    def get_users(self, role=None):
        query = Q(
            projectpermission__project=self,
            projectpermission__is_active=True,
        )
        if role:
            query = query & Q(projectpermission__role=role)

        return get_user_model().objects.filter(query).order_by('username')

    @transaction.atomic()
    def _soft_delete(self, using=None):
        """ Method for project soft delete. It doesn't delete a project, only mark as 'removed',
        but it sends needed signals and delete ServiceProjectLink objects
        """
        signals.pre_delete.send(sender=self.__class__, instance=self, using=using)

        for model in ServiceProjectLink.get_all_models():
            for spl in model.objects.filter(project=self):
                spl.delete()

        self.is_removed = True
        self.save(using=using)

        signals.post_delete.send(sender=self.__class__, instance=self, using=using)

    def delete(self, using=None, soft=True, *args, **kwargs):
        """Use soft delete, i.e. mark a project as 'removed'."""
        if soft:
            self._soft_delete(using)
        else:
            return super(SoftDeletableModel, self).delete(using=using, *args, **kwargs)

    def __str__(self):
        return '%(name)s | %(customer)s' % {
            'name': self.name,
            'customer': self.customer.name
        }

    def can_user_update_quotas(self, user):
        return user.is_staff or self.customer.has_user(user, CustomerRole.OWNER)

    def can_manage_role(self, user, role, timestamp=False):
        """
        Checks whether user can grant/update/revoke project permissions for specific role.
        `timestamp` can have following values:
            - False - check whether user can manage permissions at the moment.
            - None - check whether user can permanently manage permissions.
            - Datetime object - check whether user will be able to manage permissions at specific timestamp.
        """
        if user.is_staff:
            return True
        if self.customer.has_user(user, CustomerRole.OWNER, timestamp):
            return True

        return role == ProjectRole.ADMINISTRATOR and self.has_user(user, ProjectRole.MANAGER, timestamp)

    def get_log_fields(self):
        return ('uuid', 'customer', 'name')

    def get_parents(self):
        return [self.customer]

    def get_children(self):
        """
        Get all service project links connected to current project
        """
        return itertools.chain.from_iterable(
            m.objects.filter(project=self) for m in ServiceProjectLink.get_all_models())

    class Meta:
        base_manager_name = 'objects'


class ServiceCertification(core_models.UuidMixin, core_models.DescribableMixin):
    link = models.URLField(max_length=255, blank=True)
    # NameMixin is not used here as name has to be unique.
    name = models.CharField(_('name'), max_length=150, validators=[validate_name], unique=True)

    class Meta:
        verbose_name = 'Service Certification'
        verbose_name_plural = 'Service Certifications'
        ordering = ['-name']

    def __str__(self):
        return self.name

    @classmethod
    def get_url_name(cls):
        return 'service-certification'


class ServiceSettings(quotas_models.ExtendableQuotaModelMixin,
                      core_models.UuidMixin,
                      core_models.NameMixin,
                      core_models.StateMixin,
                      TagMixin,
                      StructureLoggableMixin):

    class Meta:
        verbose_name = "Service settings"
        verbose_name_plural = "Service settings"
        ordering = ('name',)

    class Permissions:
        customer_path = 'customer'
        extra_query = dict(shared=True)

    customer = models.ForeignKey(on_delete=models.CASCADE, to=Customer,
                                 verbose_name=_('organization'),
                                 related_name='service_settings',
                                 blank=True,
                                 null=True)
    backend_url = core_fields.BackendURLField(max_length=200, blank=True, null=True)
    username = models.CharField(max_length=100, blank=True, null=True)
    password = models.CharField(max_length=100, blank=True, null=True)
    domain = models.CharField(max_length=200, blank=True, null=True)
    token = models.CharField(max_length=255, blank=True, null=True)
    certificate = models.FileField(upload_to='certs', blank=True, null=True,
                                   validators=[CertificateValidator])
    type = models.CharField(max_length=255, db_index=True, validators=[validate_service_type])
    options = JSONField(default=dict, help_text=_('Extra options'), blank=True)
    geolocations = JSONField(default=list, blank=True,
                             help_text=_('List of latitudes and longitudes. For example: '
                                         '[{"latitude": 123, "longitude": 345}, {"latitude": 456, "longitude": 678}]'))
    shared = models.BooleanField(default=False, help_text=_('Anybody can use it'))
    homepage = models.URLField(max_length=255, blank=True)
    terms_of_services = models.URLField(max_length=255, blank=True)
    certifications = models.ManyToManyField(to='ServiceCertification', related_name='service_settings', blank=True)

    tracker = FieldTracker()

    # service settings scope - VM that contains service
    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType, null=True)
    object_id = models.PositiveIntegerField(null=True)
    scope = GenericForeignKey('content_type', 'object_id')

    objects = ServiceSettingsManager('scope')

    def get_backend(self, **kwargs):
        return SupportedServices.get_service_backend(self.type)(self, **kwargs)

    def get_option(self, name):
        options = self.options or {}
        if name in options:
            return options.get(name)
        else:
            defaults = self.get_backend().DEFAULTS
            return defaults.get(name)

    def __str__(self):
        return '%s (%s)' % (self.name, self.get_type_display())

    def get_log_fields(self):
        return ('uuid', 'name', 'customer')

    def _get_log_context(self, entity_name):
        context = super(ServiceSettings, self)._get_log_context(entity_name)
        context['service_settings_type'] = self.get_type_display()
        return context

    def get_type_display(self):
        return SupportedServices.get_name_for_type(self.type)

    def get_services(self):
        service_model = SupportedServices.get_service_models()[self.type]['service']
        return service_model.objects.filter(settings=self)

    def unlink_descendants(self):
        for service in self.get_services():
            service.unlink_descendants()
            service.delete()


class SharedServiceSettings(ServiceSettings):
    """Required for a clear separation of shared/unshared service settings on admin."""

    objects = SharedServiceSettingsManager()

    class Meta:
        proxy = True
        verbose_name_plural = _('Shared provider settings')


class PrivateServiceSettings(ServiceSettings):
    """Required for a clear separation of shared/unshared service settings on admin."""

    objects = PrivateServiceSettingsManager()

    class Meta:
        proxy = True
        verbose_name_plural = _('Private provider settings')


class Service(core_models.UuidMixin,
              core_models.DescendantMixin,
              quotas_models.QuotaModelMixin,
              StructureLoggableMixin,
              StructureModel):
    """ Base service class. """

    class Meta:
        abstract = True
        unique_together = ('customer', 'settings')

    class Permissions:
        customer_path = 'customer'
        project_path = 'projects'

    settings = models.ForeignKey(on_delete=models.CASCADE, to=ServiceSettings)
    customer = models.ForeignKey(on_delete=models.CASCADE, to=Customer, verbose_name=_('organization'))
    available_for_all = models.BooleanField(
        default=False,
        help_text=_('Service will be automatically added to all customers projects if it is available for all')
    )
    projects = NotImplemented

    def __init__(self, *args, **kwargs):
        AbstractFieldTracker().finalize_class(self.__class__, 'tracker')
        super(Service, self).__init__(*args, **kwargs)

    def __str__(self):
        return self.settings.name

    def get_backend(self, **kwargs):
        return self.settings.get_backend(**kwargs)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return cls._meta.app_label

    def get_log_fields(self):
        return ('uuid', 'customer', 'settings')

    def _get_log_context(self, entity_name):
        context = super(Service, self)._get_log_context(entity_name)
        context['service_type'] = SupportedServices.get_name_for_model(self)
        return context

    def get_service_project_links(self):
        """
        Generic method for getting queryset of service project links related to current service
        """
        return self.projects.through.objects.filter(service=self)

    def get_parents(self):
        return [self.settings, self.customer]

    def get_children(self):
        return self.get_service_project_links()

    def unlink_descendants(self):
        descendants = sort_dependencies(self._meta.model, self.get_descendants())
        for descendant in descendants:
            if isinstance(descendant, ResourceMixin):
                descendant.unlink()
            descendant.delete()


class BaseServiceProperty(core_models.BackendModelMixin, core_models.UuidMixin, core_models.NameMixin, models.Model):
    """ Base service properties like image, flavor, region,
        which are usually used for Resource provisioning.
    """

    class Meta:
        abstract = True

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return '{}-{}'.format(cls._meta.app_label, cls.__name__.lower())

    @classmethod
    def get_backend_fields(cls):
        return super(BaseServiceProperty, cls).get_backend_fields() + ('backend_id', 'name')


class ServiceProperty(BaseServiceProperty):

    class Meta:
        abstract = True
        unique_together = ('settings', 'backend_id')

    settings = models.ForeignKey(on_delete=models.CASCADE, to=ServiceSettings, related_name='+')
    backend_id = models.CharField(max_length=255, db_index=True)

    def __str__(self):
        return '{0} | {1}'.format(self.name, self.settings)


class GeneralServiceProperty(BaseServiceProperty):
    """
    Service property which is not connected to settings
    """

    class Meta:
        abstract = True

    backend_id = models.CharField(max_length=255, unique=True)

    def __str__(self):
        return self.name


class ServiceProjectLink(quotas_models.QuotaModelMixin,
                         core_models.DescendantMixin,
                         LoggableMixin,
                         StructureModel):
    """ Base service-project link class. See Service class for usage example. """

    class States:
        OK = 'OK'
        ERRED = 'ERRED'
        WARNING = 'WARNING'

        CHOICES = [OK, ERRED, WARNING]

    class Meta:
        abstract = True
        unique_together = ('service', 'project')

    class Permissions:
        customer_path = 'service__customer'
        project_path = 'project'

    service = NotImplemented
    project = models.ForeignKey(on_delete=models.CASCADE, to=Project)

    def get_backend(self, **kwargs):
        return self.service.get_backend(**kwargs)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return cls._meta.app_label + '-spl'

    def get_log_fields(self):
        return ('project', 'service',)

    def get_parents(self):
        return [self.project, self.service]

    def get_children(self):
        resource_models = [m for m in ResourceMixin.get_all_models() + SubResource.get_all_models()
                           if m.service_project_link.field.related_model == self.__class__]
        return itertools.chain.from_iterable(
            m.objects.filter(service_project_link=self) for m in resource_models)

    @property
    def validation_state(self):
        """
        Defines whether a  service compliant with required project certifications.
        """
        if set(self.project.certifications.all()).issubset(set(self.service.settings.certifications.all())):
            return self.States.OK
        else:
            return self.States.ERRED

    @property
    def is_valid(self):
        return self.validation_state == self.States.OK

    @property
    def validation_message(self):
        """
        Validation result clarification.
        """
        if not self.is_valid:
            service_certifications = self.service.settings.certifications.all()
            project_certifications = self.project.certifications.all()
            missing_certifications = set(project_certifications) - set(service_certifications)
            return _('Provider does not match with project\'s security policy. Certifications are missing: "%s"') % ', '.join([c.name for c in missing_certifications])
        else:
            return ''

    def __str__(self):
        return '{0} | {1}'.format(self.service.settings.name, self.project.name)


class CloudServiceProjectLink(ServiceProjectLink):
    """
    Represents a link between a project and a cloud service that provides VPS or VPC (e.g. Amazon, DO, OpenStack).
    """

    class Meta(ServiceProjectLink.Meta):
        abstract = True

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        vcpu = quotas_fields.QuotaField()
        ram = quotas_fields.QuotaField()
        storage = quotas_fields.QuotaField()

    def can_user_update_quotas(self, user):
        return user.is_staff or self.service.customer.has_user(user, CustomerRole.OWNER)


class ApplicationMixin(models.Model):

    class Meta:
        abstract = True

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]


class ResourceMixin(MonitoringModelMixin,
                    core_models.UuidMixin,
                    core_models.DescribableMixin,
                    core_models.NameMixin,
                    core_models.DescendantMixin,
                    core_models.BackendModelMixin,
                    StructureLoggableMixin,
                    TagMixin,
                    TimeStampedModel,
                    StructureModel):

    """ Base resource class. Resource is a provisioned entity of a service,
        for example: a VM in OpenStack or AWS, or a repository in Github.
    """

    class Meta:
        abstract = True

    class Permissions:
        customer_path = 'service_project_link__project__customer'
        project_path = 'service_project_link__project'
        service_path = 'service_project_link__service'

    service_project_link = NotImplemented
    backend_id = models.CharField(max_length=255, blank=True)

    @classmethod
    def get_backend_fields(cls):
        return super(ResourceMixin, cls).get_backend_fields() + ('backend_id',)

    def get_backend(self, **kwargs):
        return self.service_project_link.get_backend(**kwargs)

    def get_cost(self, start_date, end_date):
        raise NotImplementedError(
            "Please refer to waldur_core.billing.tasks.debit_customers while implementing it")

    def get_access_url(self):
        # default behaviour. Override in subclasses if applicable
        return None

    def get_access_url_name(self):
        return None

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls) and not issubclass(model, SubResource)]

    @classmethod
    def get_url_name(cls):
        """ This name will be used by generic relationships to membership model for URL creation """
        return '{}-{}'.format(cls._meta.app_label, cls.__name__.lower())

    def get_log_fields(self):
        return ('uuid', 'name', 'service_project_link', 'full_name')

    @property
    def full_name(self):
        return '%s %s' % (SupportedServices.get_name_for_model(self).replace('.', ' '), self.name)

    def _get_log_context(self, entity_name):
        context = super(ResourceMixin, self)._get_log_context(entity_name)
        # XXX: Add resource_full_name here, because event context does not support properties as fields
        context['resource_full_name'] = self.full_name
        context['resource_type'] = SupportedServices.get_name_for_model(self)

        return context

    def get_parents(self):
        return [self.service_project_link]

    def __str__(self):
        return self.name

    def increase_backend_quotas_usage(self, validate=True):
        """ Increase usage of quotas that were consumed by resource on creation.

            If validate is True - method should raise QuotaValidationError if
            at least one of increased quotas if over limit.
        """
        pass

    def decrease_backend_quotas_usage(self):
        """ Decrease usage of quotas that were released on resource deletion """
        pass

    def unlink(self):
        # XXX: add special attribute to an instance in order to be tracked by signal handler
        setattr(self, 'PERFORM_UNLINK', True)

    @property
    def service_settings(self):
        return self.service_project_link.service.settings

    @classmethod
    def get_scope_type(cls):
        return SupportedServices.get_name_for_model(cls)


# TODO: rename to Resource
class NewResource(ResourceMixin, core_models.StateMixin):

    class Meta:
        abstract = True


class VirtualMachine(IPCoordinatesMixin, core_models.RuntimeStateMixin, NewResource):

    def __init__(self, *args, **kwargs):
        AbstractFieldTracker().finalize_class(self.__class__, 'tracker')
        super(VirtualMachine, self).__init__(*args, **kwargs)

    cores = models.PositiveSmallIntegerField(default=0, help_text=_('Number of cores in a VM'))
    ram = models.PositiveIntegerField(default=0, help_text=_('Memory size in MiB'))
    disk = models.PositiveIntegerField(default=0, help_text=_('Disk size in MiB'))
    min_ram = models.PositiveIntegerField(default=0, help_text=_('Minimum memory size in MiB'))
    min_disk = models.PositiveIntegerField(default=0, help_text=_('Minimum disk size in MiB'))

    image_name = models.CharField(max_length=150, blank=True)

    key_name = models.CharField(max_length=50, blank=True)
    key_fingerprint = models.CharField(max_length=47, blank=True)

    user_data = models.TextField(
        blank=True,
        help_text=_('Additional data that will be added to instance on provisioning'))
    start_time = models.DateTimeField(blank=True, null=True)

    class Meta:
        abstract = True

    def detect_coordinates(self):
        if self.external_ips:
            return get_coordinates_by_ip(self.external_ips)

    def get_access_url(self):
        if self.external_ips:
            return self.external_ips
        if self.internal_ips:
            return self.internal_ips
        return None

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @property
    def external_ips(self):
        """
        Returns a list of external IPs.
        Implementation of this method in all derived classes guarantees all virtual machine have the same interface.
        For instance:
         - SaltStack (aws) handles IPs as private and public IP addresses;
         - DigitalOcean has only 1 external ip called ip_address.
        """
        return []

    @property
    def internal_ips(self):
        """
        Returns a list of internal IPs.
        Implementation of this method in all derived classes guarantees all virtual machine have the same interface.
        For instance:
         - SaltStack (aws) handles IPs as private and public IP addresses;
         - DigitalOcean does not support internal IP at the moment.
        """
        return []


class PrivateCloud(quotas_models.QuotaModelMixin, core_models.RuntimeStateMixin, NewResource):
    extra_configuration = JSONField(default=dict,
                                    help_text=_('Configuration details that are not represented on backend.'))

    class Meta:
        abstract = True


class Storage(core_models.RuntimeStateMixin, NewResource):
    size = models.PositiveIntegerField(help_text=_('Size in MiB'))

    class Meta:
        abstract = True


class Volume(Storage):
    class Meta:
        abstract = True


class Snapshot(Storage):
    class Meta:
        abstract = True


class SubResource(NewResource):
    """ Resource dependent object that cannot exist without resource. """
    class Meta:
        abstract = True

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]



from django.contrib import admin

from . import models


class AuthResultAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'state', 'user', 'modified')
    ordering = ('modified',)
    list_filter = ('state', 'user')


admin.site.register(models.AuthResult, AuthResultAdmin)



from celery import shared_task
from django.conf import settings

from waldur_mastermind.marketplace import models
from waldur_mastermind.marketplace_script import serializers, utils, PLUGIN_NAME


@shared_task(name='waldur_marketplace_script.pull_resources')
def pull_resources():
    for resource in models.Resource.objects.filter(
        offering__type=PLUGIN_NAME,
        offering__plugin_options__has_key='pull',
        state__in=[models.Resource.States.OK, models.Resource.States.ERRED]
    ):
        pull_resource.delay(resource.id)


@shared_task
def pull_resource(resource_id):
    resource = models.Resource.objects.get(id=resource_id)
    options = resource.offering.plugin_options

    serializer = serializers.ResourceSerializer(instance=resource)
    environment = {key.upper(): str(value) for key, value in serializer.data}
    if isinstance(options.get('environ'), dict):
        environment.update(options['environ'])

    language = options['language']
    image = settings.WALDUR_MARKETPLACE_SCRIPT['DOCKER_IMAGES'].get(language)
    utils.execute_script(
        image=image,
        command=language,
        src=options['pull'],
        environment=environment
    )



import json

import requests


class RijkscloudClient:
    """
    Rijkscloud Python client.
    """

    def __init__(self, apikey, userid):
        self.base_url = 'https://api.ix.rijkscloud.nl'
        self.headers = {
            'Content-Type': 'application/json',
            'apikey': apikey,
            'userid': userid,
        }

    def _get(self, endpoint, key):
        url = '%s/%s' % (self.base_url, endpoint)
        response = requests.get(url, headers=self.headers)
        response.raise_for_status()
        data = response.json()
        if key:
            return data.get(key)
        else:
            return data

    def _post(self, endpoint, body):
        url = '%s/%s' % (self.base_url, endpoint)
        response = requests.post(url, headers=self.headers, data=json.dumps(body))
        if response.status_code == 400 and response.content:
            message = response.json()['error']['message']
            raise requests.HTTPError(message, response=response)

        response.raise_for_status()
        if response.content:
            return response.json()

    def _delete(self, endpoint):
        url = '%s/%s' % (self.base_url, endpoint)
        response = requests.delete(url, headers=self.headers)
        response.raise_for_status()
        if response.content:
            return response.json()

    def list_flavors(self):
        return self._get('flavors', 'flavors')

    def get_flavor(self, flavor_name):
        flavors_map = {flavor['name']: flavor for flavor in self.list_flavors()}
        return flavors_map.get(flavor_name)

    def get_instance(self, instance_name):
        url = 'instances/%s' % instance_name
        return self._get(url, 'instance')

    def list_instances(self):
        instances = self._get('instances', 'instances')
        return [self.get_instance(instance['name']) for instance in instances]

    def create_instance(self, body):
        return self._post('instances', body)

    def delete_instance(self, instance_name):
        return self._delete('instances/%s' % instance_name)

    def get_network(self, network_name):
        subnets = self.list_subnets(network_name)
        return dict(name=network_name, subnets=subnets)

    def list_networks(self):
        networks = self._get('networks', 'networks')
        return [self.get_network(network['name']) for network in networks]

    def list_floatingips(self):
        return self._get('networks/floats', 'floats')

    def get_subnet(self, network_name, subnet_name):
        url = 'networks/%s/subnets/%s' % (network_name, subnet_name)
        subnet = self._get(url, 'subnet')
        ips = self.list_subnet_ips(network_name, subnet_name)
        return dict(name=subnet_name, ips=ips, **subnet)

    def list_subnet_ips(self, network_name, subnet_name):
        url = 'networks/%s/subnets/%s/ips' % (network_name, subnet_name)
        return self._get(url, 'ips')

    def list_subnets(self, network_name):
        url = 'networks/%s/subnets' % network_name
        subnets = self._get(url, 'subnets')
        return [self.get_subnet(network_name, subnet['name']) for subnet in subnets]

    def get_volume(self, volume_name):
        url = 'volumes/%s' % volume_name
        return self._get(url, 'volume')

    def list_volumes(self):
        volumes = self._get('volumes', 'volumes')
        return [self.get_volume(volume['name']) for volume in volumes]

    def create_volume(self, body):
        return self._post('volumes', body)

    def delete_volume(self, volume_name):
        return self._delete('volumes/%s' % volume_name)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-27 09:50
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('logging', '0006_event_feed'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='alert',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='alert',
            name='content_type',
        ),
        migrations.DeleteModel(
            name='Alert',
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.core import serializers as django_serializers
from django.core.management.base import BaseCommand
from reversion.models import Version, Revision

from waldur_core.quotas import models
from waldur_core.quotas.utils import get_models_with_quotas


class Command(BaseCommand):
    """ Recalculate all quotas """

    def handle(self, *args, **options):
        for model in get_models_with_quotas():
            if hasattr(model, 'GLOBAL_COUNT_QUOTA_NAME'):
                quota, _ = models.Quota.objects.get_or_create(name=model.GLOBAL_COUNT_QUOTA_NAME)
                for index, instance in enumerate(model.objects.all().order_by('created')):
                    revision = Revision.objects.create()
                    revision.date_created = instance.created
                    revision.save()

                    quota.usage = index + 1
                    serializer = django_serializers.get_serializer('json')()
                    serialized_data = serializer.serialize([quota])

                    Version.objects.create(
                        revision=revision,
                        object_id=quota.id,
                        object_id_int=quota.id,
                        content_type=ContentType.objects.get_for_model(quota),
                        format='json',
                        serialized_data=serialized_data,
                        object_repr=str(quota),
                    )



import logging
from celery import shared_task

from waldur_core.core import utils as core_utils

from . import backend, exceptions

logger = logging.getLogger(__name__)


@shared_task
def create_doi(serialized_instance):
    instance = core_utils.deserialize_instance(serialized_instance)

    if instance.datacite_doi:
        logger.warning(
            'Registration of %s has been skipped because datacite_doi field is not empty.' % instance)
        return

    try:
        backend.DataciteBackend().create_doi(instance)
    except exceptions.DataciteException as e:
        logger.critical(e)



# Generated by Django 2.2.7 on 2019-12-16 11:15

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0008_remove_offering_enable_dynamic_components'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.Offering'),
        ),
    ]



import logging

from celery import shared_task
from django.core import exceptions

from waldur_core.core import utils
from waldur_geo_ip import utils as geo_ip_utils

logger = logging.getLogger(__name__)


@shared_task(name='waldur_geo_ip.detect_vm_coordinates')
def detect_vm_coordinates(serialized_virtual_machine):

    try:
        vm = utils.deserialize_instance(serialized_virtual_machine)
    except exceptions.ObjectDoesNotExist:
        logger.warning('Missing virtual machine %s.', serialized_virtual_machine)
        return

    geo_ip_utils.detect_coordinates(vm)


@shared_task(name='waldur_geo_ip.detect_vm_coordinates_batch')
def detect_vm_coordinates_batch(serialized_virtual_machines):
    for vm in serialized_virtual_machines:
        detect_vm_coordinates.delay(vm)



import unicodedata

from django.contrib.auth import password_validation
from django.contrib.auth.hashers import (
    check_password, is_password_usable, make_password,
)
from django.db import models
from django.db.models.fields.related import lazy_related_operation
from django.utils.crypto import salted_hmac
from django.utils.translation import gettext_lazy as _
from taggit.managers import TaggableManager as _TaggableManager


class AbstractBaseUser(models.Model):
    """
    is_active is removed
    """
    password = models.CharField(_('password'), max_length=128)
    last_login = models.DateTimeField(_('last login'), blank=True, null=True)

    REQUIRED_FIELDS = []

    # Stores the raw password if set_password() is called so that it can
    # be passed to password_changed() after the model is saved.
    _password = None

    class Meta:
        abstract = True

    def __str__(self):
        return self.get_username()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self._password is not None:
            password_validation.password_changed(self._password, self)
            self._password = None

    def get_username(self):
        """Return the username for this User."""
        return getattr(self, self.USERNAME_FIELD)

    def clean(self):
        setattr(self, self.USERNAME_FIELD, self.normalize_username(self.get_username()))

    def natural_key(self):
        return (self.get_username(),)

    @property
    def is_anonymous(self):
        """
        Always return False. This is a way of comparing User objects to
        anonymous users.
        """
        return False

    @property
    def is_authenticated(self):
        """
        Always return True. This is a way to tell if the user has been
        authenticated in templates.
        """
        return True

    def set_password(self, raw_password):
        self.password = make_password(raw_password)
        self._password = raw_password

    def check_password(self, raw_password):
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """
        def setter(raw_password):
            self.set_password(raw_password)
            # Password hash upgrades shouldn't be considered password changes.
            self._password = None
            self.save(update_fields=["password"])
        return check_password(raw_password, self.password, setter)

    def set_unusable_password(self):
        # Set a value that will never be a valid hash
        self.password = make_password(None)

    def has_usable_password(self):
        """
        Return False if set_unusable_password() has been called for this user.
        """
        return is_password_usable(self.password)

    def get_session_auth_hash(self):
        """
        Return an HMAC of the password field.
        """
        key_salt = "django.contrib.auth.models.AbstractBaseUser.get_session_auth_hash"
        return salted_hmac(key_salt, self.password).hexdigest()

    @classmethod
    def get_email_field_name(cls):
        try:
            return cls.EMAIL_FIELD
        except AttributeError:
            return 'email'

    @classmethod
    def normalize_username(cls, username):
        return unicodedata.normalize('NFKC', username) if isinstance(username, str) else username


class TaggableManager(_TaggableManager):
    """
    Modify contribute_to_class method, so it can use in abstract class
    See also: https://github.com/jazzband/django-taggit/pull/632/files
    """
    def contribute_to_class(self, cls, name):
        self.set_attributes_from_name(name)
        self.model = cls
        self.opts = cls._meta

        cls._meta.add_field(self)
        setattr(cls, name, self)
        if not cls._meta.abstract:
            self.remote_field.related_name = '%(class)s_%(model_name)s_%(app_label)s' % {
                "class": cls.__name__.lower(),
                "model_name": cls._meta.model_name.lower(),
                "app_label": cls._meta.app_label.lower(),
            }

            if self.remote_field.related_query_name:
                related_query_name = self.remote_field.related_query_name % {
                    "class": cls.__name__.lower(),
                    "app_label": cls._meta.app_label.lower(),
                }
                self.remote_field.related_query_name = related_query_name

            if isinstance(self.remote_field.model, str):

                def resolve_related_class(cls, model, field):
                    field.remote_field.model = model

                lazy_related_operation(
                    resolve_related_class, cls, self.remote_field.model, field=self
                )
            if isinstance(self.through, str):

                def resolve_related_class(cls, model, field):
                    self.through = model
                    self.remote_field.through = model
                    self.post_through_setup(cls)

                lazy_related_operation(
                    resolve_related_class, cls, self.through, field=self
                )
            else:
                self.post_through_setup(cls)



# Generated by Django 2.2.7 on 2019-12-03 17:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0002_nullable_issue_caller'),
    ]

    operations = [
        migrations.AlterField(
            model_name='supportcustomer',
            name='user',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



# -*- coding: utf-8 -*-

import django.utils.timezone
import model_utils.fields
from django.conf import settings
from django.db import migrations, models

import waldur_core.core.fields
import waldur_core.structure.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0001_squashed_0054'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Invitation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False,
                                                                verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False,
                                                                      verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('customer_role', waldur_core.structure.models.CustomerRole(blank=True, choices=[('owner', 'Owner'), (
                'support', 'Support')], max_length=30, null=True, verbose_name='organization role')),
                ('project_role', waldur_core.structure.models.ProjectRole(blank=True,
                                                                          choices=[('admin', 'Administrator'),
                                                                                   ('manager', 'Manager'),
                                                                                   ('support', 'Support')],
                                                                          max_length=30, null=True)),
                ('state', models.CharField(
                    choices=[('accepted', 'Accepted'), ('canceled', 'Canceled'), ('pending', 'Pending'),
                             ('expired', 'Expired')], default='pending', max_length=8)),
                ('link_template', models.CharField(
                    help_text='The template must include {uuid} parameter e.g. http://example.com/invitation/{uuid}',
                    max_length=255)),
                ('email', models.EmailField(
                    help_text='Invitation link will be sent to this email. Note that user can accept invitation with different email.',
                    max_length=254)),
                ('civil_number', models.CharField(blank=True,
                                                  help_text='Civil number of invited user. If civil number is not defined any user can accept invitation.',
                                                  max_length=50)),
                ('created_by',
                 models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+',
                                   to=settings.AUTH_USER_MODEL)),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='invitations',
                                               to='structure.Customer', verbose_name='organization')),
                ('project', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                              related_name='invitations', to='structure.Project')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from rest_framework import serializers

from waldur_core.core.serializers import Base64Field
from . import models, utils


class Saml2LoginSerializer(serializers.Serializer):
    idp = serializers.CharField()

    def clean_idp(self, value):
        if utils.is_valid_idp(value):
            return value
        else:
            raise serializers.ValidationError('Identity provider %s is not available.' % value)


class Saml2LoginCompleteSerializer(serializers.Serializer):
    SAMLResponse = Base64Field()


class Saml2LogoutCompleteSerializer(serializers.Serializer):
    SAMLResponse = Base64Field(required=False)
    SAMLRequest = Base64Field(required=False)

    def validate(self, attrs):
        if not attrs.get('SAMLResponse') and not attrs.get('SAMLRequest'):
            raise serializers.ValidationError('Either SAMLResponse or SAMLRequest must be provided.')

        return attrs


class Saml2ProviderSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.IdentityProvider
        fields = ('name', 'url')



from django.db import models
from django.utils.translation import ugettext_lazy as _
from waldur_core.core.fields import JSONField
from waldur_core.logging.loggers import LoggableMixin
from waldur_core.structure import models as structure_models


class RijkscloudService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project,
        related_name='rijkscloud_services',
        through='RijkscloudServiceProjectLink'
    )

    class Meta:
        unique_together = ('customer', 'settings')
        verbose_name = _('Rijkscloud provider')
        verbose_name_plural = _('Rijkscloud providers')

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud'


class RijkscloudServiceProjectLink(structure_models.ServiceProjectLink):

    service = models.ForeignKey(on_delete=models.CASCADE, to=RijkscloudService)

    class Meta(structure_models.ServiceProjectLink.Meta):
        verbose_name = _('Rijkscloud provider project link')
        verbose_name_plural = _('Rijkscloud provider project links')

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-spl'


class Flavor(LoggableMixin, structure_models.ServiceProperty):
    cores = models.PositiveSmallIntegerField(help_text=_('Number of cores in a VM'))
    ram = models.PositiveIntegerField(help_text=_('Memory size in MiB'))

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-flavor'

    @classmethod
    def get_backend_fields(cls):
        readonly_fields = super(Flavor, cls).get_backend_fields()
        return readonly_fields + ('cores', 'ram')


class Volume(structure_models.Volume):
    service_project_link = models.ForeignKey(
        RijkscloudServiceProjectLink,
        related_name='volumes',
        on_delete=models.PROTECT
    )
    metadata = JSONField(blank=True)

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-volume'

    @classmethod
    def get_backend_fields(cls):
        return super(Volume, cls).get_backend_fields() + (
            'name', 'size', 'metadata', 'runtime_state')


class Instance(structure_models.VirtualMachine):
    service_project_link = models.ForeignKey(
        RijkscloudServiceProjectLink,
        related_name='instances',
        on_delete=models.PROTECT
    )
    flavor_name = models.CharField(max_length=255, blank=True)
    floating_ip = models.ForeignKey(on_delete=models.CASCADE, to='FloatingIP', blank=True, null=True)
    internal_ip = models.ForeignKey(on_delete=models.CASCADE, to='InternalIP')

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-instance'

    @classmethod
    def get_backend_fields(cls):
        return super(Instance, cls).get_backend_fields() + (
            'flavor_name', 'ram', 'cores', 'runtime_state'
        )

    @property
    def external_ips(self):
        if self.floating_ip:
            return [self.floating_ip.address]

    @property
    def internal_ips(self):
        return [self.internal_ip.address]

    def __str__(self):
        return self.name


class FloatingIP(structure_models.ServiceProperty):
    address = models.GenericIPAddressField(protocol='IPv4', null=True)
    is_available = models.BooleanField(default=True)

    class Meta:
        verbose_name = _('Floating IP')
        verbose_name_plural = _('Floating IPs')
        unique_together = ('settings', 'backend_id')

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-fip'

    @classmethod
    def get_backend_fields(cls):
        return super(FloatingIP, cls).get_backend_fields() + ('address', 'is_available')

    def __str__(self):
        return self.address


class Network(structure_models.ServiceProperty):
    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-network'

    def __str__(self):
        return self.name


class SubNet(structure_models.ServiceProperty):
    network = models.ForeignKey(on_delete=models.CASCADE, to=Network, related_name='subnets')
    cidr = models.CharField(max_length=32)
    gateway_ip = models.GenericIPAddressField(protocol='IPv4')
    allocation_pools = JSONField()
    dns_nameservers = JSONField(help_text=_('List of DNS name servers associated with the subnet.'))

    class Meta:
        verbose_name = _('Subnet')
        verbose_name_plural = _('Subnets')
        unique_together = ('settings', 'backend_id')

    def __str__(self):
        return '%s (%s)' % (self.name, self.cidr)

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-subnet'


class InternalIP(structure_models.ServiceProperty):
    address = models.GenericIPAddressField(protocol='IPv4')
    is_available = models.BooleanField(default=True)
    subnet = models.ForeignKey(on_delete=models.CASCADE, to=SubNet, related_name='internal_ips')

    @classmethod
    def get_backend_fields(cls):
        return super(InternalIP, cls).get_backend_fields() + ('address', 'is_available')

    @classmethod
    def get_url_name(cls):
        return 'rijkscloud-internal-ip'

    class Meta:
        verbose_name = _('Internal IP')
        verbose_name_plural = _('Internal IPs')
        unique_together = ('settings', 'backend_id')

    def __str__(self):
        return '%s (%s)' % (self.address, self.subnet.name)



from waldur_core.logging.loggers import EventLogger, event_logger


class DropletResizeEventLogger(EventLogger):
    droplet = 'waldur_digitalocean.Droplet'
    size = 'waldur_digitalocean.Size'

    class Meta:
        event_types = ('droplet_resize_scheduled',
                       'droplet_resize_succeeded')

    @staticmethod
    def get_scopes(event_context):
        resource = event_context['droplet']
        project = resource.service_project_link.project
        return {resource, project, project.customer}


event_logger.register('droplet_resize', DropletResizeEventLogger)



from waldur_core.core import WaldurExtension


class PackagesExtension(WaldurExtension):

    class Settings:
        WALDUR_PACKAGES = {
            'BILLING_ENABLED': False,
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.packages'

    @staticmethod
    def is_assembly():
        return True



from django.conf import settings
from rest_framework import views
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from . import client, serializers


class UserDetailsViewSet(views.APIView):

    def get(self, request, *args, **kwargs):
        if not settings.WALDUR_AUTH_BCC['ENABLED']:
            raise ValidationError('This feature is disabled.')

        serializer = serializers.UserDetailRequestSerializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)

        civil_number = serializer.validated_data['civil_number']
        tax_number = serializer.validated_data['tax_number']

        try:
            user_details = client.get_user_details(civil_number, tax_number)
            return Response(user_details._asdict())
        except client.BCCException as e:
            return Response({'details': e.detail}, status=e.code)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2018-03-02 11:41
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0013_extend_jira_issue_fields'),
    ]

    operations = [
        migrations.RenameField(
            model_name='issue',
            old_name='updated_username',
            new_name='creator_username',
        ),
        migrations.RemoveField(
            model_name='issue',
            name='assignee',
        ),
        migrations.AddField(
            model_name='issue',
            name='assignee_email',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='issue',
            name='assignee_name',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='issue',
            name='assignee_username',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='issue',
            name='creator_email',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='issue',
            name='reporter_email',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='issue',
            name='reporter_name',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='issue',
            name='reporter_username',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.utils import timezone

from waldur_core.cost_tracking.models import PriceEstimate
from waldur_core.structure.models import ServiceSettings


def get_total_cost():
    points = []
    estimates, mapping = get_current_estimates()
    for estimate in estimates.only('consumed', 'object_id'):
        points.append({
            'measurement': 'total_cost',
            'tags': {
                'provider': mapping[estimate.object_id],
            },
            'fields': {
                'value': estimate.consumed,
            },
        })
    return points


def get_current_estimates():
    content_type = ContentType.objects.get_for_model(ServiceSettings)
    service_settings = ServiceSettings.objects.filter(shared=True).only('pk', 'name')
    mapping = {item.id: item.name for item in service_settings}
    now = timezone.now()
    estimates = PriceEstimate.objects.filter(
        content_type=content_type,
        object_id__in=mapping.keys(),
        month=now.month,
        year=now.year,
    )
    return estimates, mapping



from django.apps import AppConfig
from django.db.models import signals


class OpenStackConfig(AppConfig):
    """ OpenStack is a toolkit for building private and public clouds.
        This application adds support for managing OpenStack deployments -
        tenants, instances, security groups and networks.
    """
    name = 'waldur_openstack.openstack'
    label = 'openstack'
    verbose_name = 'OpenStack'
    service_name = 'OpenStack'

    def ready(self):
        from waldur_core.core import models as core_models
        from waldur_core.quotas.fields import QuotaField
        from waldur_core.structure import models as structure_models
        from waldur_core.structure import signals as structure_signals
        from waldur_core.structure import SupportedServices
        from . import handlers

        Tenant = self.get_model('Tenant')

        # structure
        from .backend import OpenStackBackend
        SupportedServices.register_backend(OpenStackBackend)

        from . import quotas
        quotas.inject_tenant_quotas()

        for resource in ('vcpu', 'ram', 'storage'):
            structure_models.ServiceSettings.add_quota_field(
                name='openstack_%s' % resource,
                quota_field=QuotaField(
                    creation_condition=lambda service_settings:
                        service_settings.type == OpenStackConfig.service_name
                )
            )

        for model in (structure_models.Project, structure_models.Customer):
            structure_signals.structure_role_revoked.connect(
                handlers.remove_ssh_key_from_tenants,
                sender=model,
                dispatch_uid='openstack.handlers.remove_ssh_key_from_tenants__%s' % model.__name__,
            )

        signals.pre_delete.connect(
            handlers.remove_ssh_key_from_all_tenants_on_it_deletion,
            sender=core_models.SshPublicKey,
            dispatch_uid='openstack.handlers.remove_ssh_key_from_all_tenants_on_it_deletion',
        )

        from waldur_core.quotas.models import Quota
        signals.post_save.connect(
            handlers.log_tenant_quota_update,
            sender=Quota,
            dispatch_uid='openstack.handlers.log_tenant_quota_update',
        )

        signals.post_save.connect(
            handlers.update_service_settings_name,
            sender=Tenant,
            dispatch_uid='openstack.handlers.update_service_settings_name',
        )



import os

from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import ugettext_lazy as _

from waldur_core.media import magic

# max bytes to read for file type detection
READ_SIZE = 5 * (1024 * 1024)   # 5MB


# Based on https://github.com/mckinseyacademy/django-upload-validator/blob/master/upload_validator/__init__.py
@deconstructible
class FileTypeValidator:
    """
    File type validator for validating mimetypes and extensions
    Args:
        allowed_types (list): list of acceptable mimetypes e.g; ['image/jpeg', 'application/pdf']
                    see https://www.iana.org/assignments/media-types/media-types.xhtml
        allowed_extensions (list, optional): list of allowed file extensions e.g; ['.jpeg', '.pdf', '.docx']
    """
    type_message = _(
        "File type '%(detected_type)s' is not allowed. "
        "Allowed types are: '%(allowed_types)s'."
    )

    extension_message = _(
        "File extension '%(extension)s' is not allowed. "
        "Allowed extensions are: '%(allowed_extensions)s'."
    )

    def __init__(self, allowed_types, allowed_extensions=()):
        self.allowed_mimes = allowed_types
        self.allowed_exts = allowed_extensions

    def __call__(self, fileobj):
        detected_type = magic.from_buffer(fileobj.read(READ_SIZE), mime=True)
        root, extension = os.path.splitext(fileobj.name.lower())

        # seek back to start so a valid file could be read
        # later without resetting the position
        fileobj.seek(0)

        # some versions of libmagic do not report proper mimes for Office subtypes
        # use detection details to transform it to proper mime
        if detected_type in ('application/octet-stream', 'application/vnd.ms-office'):
            detected_type = self.check_word_or_excel(fileobj, detected_type, extension)

        if detected_type not in self.allowed_mimes:
            # use more readable file type names for feedback message
            allowed_types = map(lambda mime_type: mime_type.split('/')[1], self.allowed_mimes)

            raise ValidationError(
                message=self.type_message,
                params={
                    'detected_type': detected_type,
                    'allowed_types': ', '.join(allowed_types)
                },
                code='invalid_type'
            )

        if self.allowed_exts and (extension not in self.allowed_exts):
            raise ValidationError(
                message=self.extension_message,
                params={
                    'extension': extension,
                    'allowed_extensions': ', '.join(self.allowed_exts)
                },
                code='invalid_extension'
            )

    def check_word_or_excel(self, fileobj, detected_type, extension):
        """
        Returns proper mimetype in case of word or excel files
        """
        word_strings = ['Microsoft Word', 'Microsoft Office Word', 'Microsoft Macintosh Word']
        excel_strings = ['Microsoft Excel', 'Microsoft Office Excel', 'Microsoft Macintosh Excel']
        office_strings = ['Microsoft OOXML']

        file_type_details = magic.from_buffer(fileobj.read(READ_SIZE))

        fileobj.seek(0)

        if any(string in file_type_details for string in word_strings):
            detected_type = 'application/msword'
        elif any(string in file_type_details for string in excel_strings):
            detected_type = 'application/vnd.ms-excel'
        elif any(string in file_type_details for string in office_strings) or \
                (detected_type == 'application/vnd.ms-office'):
            if extension in ('.doc', '.docx'):
                detected_type = 'application/msword'
            if extension in ('.xls', '.xlsx'):
                detected_type = 'application/vnd.ms-excel'

        return detected_type


ImageValidator = FileTypeValidator(
    allowed_types=[
        'image/png',
        'image/gif',
        'image/jpeg',
        'image/svg',
        'image/svg+xml',
        'image/x-icon',
    ]
)


CertificateValidator = FileTypeValidator(
    allowed_types=[
        'application/x-pem-file',
        'application/x-x509-ca-cert',
        'text/plain'
    ],
    allowed_extensions=['pem']
)


DocumentValidator = FileTypeValidator(
    allowed_types=[
        'text/csv',
        'text/html',
        'text/plain',
        'application/xhtml+xml',
        'application/pdf',
        'application/rtf',
        'application/msword',
        'application/vnd.ms-office',
        'application/vnd.ms-powerpoint',
        'application/vnd.ms-excel',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        'application/vnd.openxmlformats-officedocument.presentationml.presentation',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        'application/vnd.oasis.opendocument.text',
        'application/vnd.oasis.opendocument.spreadsheet',
        'application/vnd.oasis.opendocument.presentation'
    ],
    allowed_extensions=[
        'csv',
        'txt',
        'htm',
        'html',
        'xhtml',
        'pdf',
        'rtf',
        'doc',
        'xls',
        'ppt',
        'docx',
        'xlsx',
        'pptx',
        'odt',
        'odp',
        'ods',
        'rtf',
    ]
)



from rest_framework import serializers

from waldur_core.core.serializers import GenericRelatedField
from waldur_core.quotas import models, utils


class QuotaSerializer(serializers.HyperlinkedModelSerializer):
    scope = GenericRelatedField(related_models=utils.get_models_with_quotas(), read_only=True)

    class Meta:
        model = models.Quota
        fields = ('url', 'uuid', 'name', 'limit', 'usage', 'scope', 'threshold')
        read_only_fields = ('uuid', 'name', 'usage')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class BasicQuotaSerializer(serializers.HyperlinkedModelSerializer):
    """
    It does not expose scope in order to reduce number of queries
    """
    class Meta:
        model = models.Quota
        fields = ('url', 'uuid', 'name', 'limit', 'usage')
        read_only_fields = ('uuid', 'name', 'usage')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }



from django import http


def cors_middleware(get_response):
    """
    If CORS preflight header, then create an empty body response (200 OK) and return it
    """

    def middleware(request):
        if request.method == "OPTIONS" and "HTTP_ACCESS_CONTROL_REQUEST_METHOD" in request.META:
            response = http.HttpResponse()
            response["Content-Length"] = "0"
            return response

        return get_response(request)

    return middleware



from django.db import migrations


PACKAGE_TYPE = 'Packages.Template'

STORAGE_MODE_FIXED = 'fixed'


def set_storage_mode(apps, schema_editor):
    Offering = apps.get_model('marketplace', 'Offering')
    Offering.objects.filter(type=PACKAGE_TYPE).update(plugin_options={'storage_mode': STORAGE_MODE_FIXED})


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace_openstack', '0002_prefix_storage'),
    ]

    operations = [
        migrations.RunPython(set_storage_mode)
    ]



# Generated by Django 2.2.7 on 2019-12-27 10:29

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_rancher', '0007_cluster_tenant_settings'),
    ]

    operations = [
        migrations.AddField(
            model_name='node',
            name='annotations',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict),
        ),
        migrations.AddField(
            model_name='node',
            name='cpu_allocated',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='node',
            name='cpu_total',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='node',
            name='docker_version',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='node',
            name='k8s_version',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='node',
            name='labels',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict),
        ),
        migrations.AddField(
            model_name='node',
            name='pods_allocated',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='node',
            name='pods_total',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='node',
            name='ram_allocated',
            field=models.IntegerField(blank=True, null=True, help_text='Allocated RAM in Mi.'),
        ),
        migrations.AddField(
            model_name='node',
            name='ram_total',
            field=models.IntegerField(blank=True, null=True, help_text='Total RAM in Mi.'),
        ),
    ]



from rest_framework import response, status, views

from . import estimator


class AnsibleEstimatorView(views.APIView):
    def post(self, request, format=None):
        report = estimator.get_report(request)
        return response.Response(report, status=status.HTTP_200_OK)



import datetime
from urllib.parse import urlencode, urlparse, parse_qs
from urllib.request import urlopen

import dateutil.parser
import decimal
import paypalrestsdk as paypal

from django.conf import settings
from django.core.files.base import ContentFile
from django.utils import timezone


class PayPalError(Exception):
    pass


class PaypalPayment:
    def __init__(self, payment_id, approval_url, token):
        self.payment_id = payment_id
        self.approval_url = approval_url
        self.token = token


class PaypalBackend:

    BACKEND_SERVERS_MAP = {
        'sandbox': 'https://www.sandbox.paypal.com',
        'live': 'https://www.paypal.com',
    }

    def __init__(self):
        config = settings.WALDUR_PAYPAL['BACKEND']
        self.configure(**config)

    def configure(self, mode, client_id, client_secret, currency_name, **kwargs):
        # extra method to validate required config options
        self.currency_name = currency_name

        paypal.configure({
            'mode': mode,
            'client_id': client_id,
            'client_secret': client_secret
        })

        self.server = self.BACKEND_SERVERS_MAP[mode]

    def get_payment_view_url(self, backend_invoice_id, params=None):
        invoice_url = '%s/invoice/payerView/details/%s' % (self.server, backend_invoice_id)
        if params:
            query_params = urlencode(params)
            invoice_url = '%s?%s' % (invoice_url, query_params)

        return invoice_url

    def _find_approval_url(self, links):
        for link in links:
            if link.rel == 'approval_url':
                return link.href
        raise PayPalError('Approval URL is not found')

    def _find_token(self, approval_url):
        parts = urlparse(approval_url)
        params = parse_qs(parts.query)
        token = params.get('token')
        if not token:
            raise PayPalError('Unable to parse token from approval_url')
        return token[0]

    def create_invoice(self, invoice):
        """
        Creates invoice with invoice items in it.
        https://developer.paypal.com/docs/api/invoicing/#definition-payment_term
        :param invoice: instance of Invoice class.
        :return: instance of Invoice with backend_id filled.
        """

        if invoice.backend_id:
            return

        if not invoice.items.count():
            raise PayPalError('"items" size must be between 1 and 100.')

        if not invoice.price:
            raise PayPalError('The total cost must not be zero.')

        phone = invoice.issuer_details.get('phone', {})
        if not phone:
            raise PayPalError('"phone" is a required attribute')

        if phone and 'country_code' not in phone:
            raise PayPalError('"phone"."country_code" is a required attribute')

        if phone and 'national_number' not in phone:
            raise PayPalError('"phone"."national_number" is a required attribute')

        invoice_details = {
            'merchant_info': {
                'email': invoice.issuer_details.get('email'),
                'business_name': invoice.issuer_details.get('company'),
                'phone': {
                    'country_code': phone.get('country_code'),
                    'national_number': phone.get('national_number'),
                },
                'address': {
                    'line1': invoice.issuer_details.get('address'),
                    'city': invoice.issuer_details.get('city'),
                    'state': invoice.issuer_details.get('state'),
                    'postal_code': invoice.issuer_details.get('postal'),
                    'country_code': invoice.issuer_details.get('country_code')
                }
            },
            'items': [
                {
                    'name': item.name,
                    'unit_of_measure': item.unit_of_measure,
                    'quantity': item.quantity,
                    'date': self._format_date(item.start.date()),
                    'unit_price': {
                        'currency': self.currency_name,
                        'value': self._format_decimal(item.unit_price),
                    }
                } for item in invoice.items.iterator()
            ],
            'tax_inclusive': False,
            'payment_term': {
                'due_date': self._format_date(invoice.end_date),
            },
            'total_amount': {
                'currency': self.currency_name,
                'value': self._format_decimal(invoice.total)
            }
            # 'logo_url': pass logo url if needed. 250x90, HTTPS. Image is not displayed o PDF atm.
        }

        if invoice.tax_percent and invoice.tax_percent > 0:
            for item in invoice_details['items']:
                item['tax'] = {
                    'name': 'VAT',
                    'percent': self._format_decimal(invoice.tax_percent),
                }

        invoice_details['billing_info'] = [
            {
                'email': invoice.customer.email,
                'business_name': invoice.customer.name,
            }
        ]

        backend_invoice = paypal.Invoice(invoice_details)

        try:
            if backend_invoice.create():
                invoice.state = backend_invoice.status
                invoice.backend_id = backend_invoice.id
                invoice.number = backend_invoice.number
                invoice.save(update_fields=['state', 'backend_id', 'number'])

                return invoice
            else:
                raise PayPalError(backend_invoice.error)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def send_invoice(self, invoice):
        if invoice.state != invoice.States.DRAFT:
            raise PayPalError('Invoice must be in "%s" state' % invoice.States.DRAFT)

        try:
            backend_invoice = paypal.Invoice.find(invoice.backend_id)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

        if not backend_invoice.send():
            raise PayPalError(backend_invoice.error)

        return invoice

    def pull_invoice(self, invoice):
        try:
            backend_invoice = paypal.Invoice.find(invoice.backend_id)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

        invoice.state = backend_invoice.status
        invoice.number = backend_invoice.number
        invoice.save(update_fields=['state', 'number'])
        return invoice

    def make_payment(self, amount, tax, description, return_url, cancel_url):
        """
        Make PayPal payment using Express Checkout workflow.
        https://developer.paypal.com/docs/api/payments/

        :param amount: Decimal value of payment including VAT tax.
        :param tax: Decimal value of VAT tax.
        :param description: Description of payment.
        :param return_url: Callback view URL for approved payment.
        :param cancel_url: Callback view URL for cancelled payment.
        :return: Object containing backend payment id, approval URL and token.
        """
        if amount < tax:
            raise PayPalError('Payment amount should be greater than tax.')

        payment = paypal.Payment({
            'intent': 'sale',
            'payer': {'payment_method': 'paypal'},
            'transactions': [
                {
                    'amount': {
                        'total': self._format_decimal(amount),
                        'currency': self.currency_name,
                        'details': {
                            'subtotal': self._format_decimal(amount - tax),
                            'tax': self._format_decimal(tax)
                        }
                    },
                    'description': description
                }
            ],
            'redirect_urls': {
                'return_url': return_url,
                'cancel_url': cancel_url
            }
        })

        try:
            if payment.create():
                approval_url = self._find_approval_url(payment.links)
                token = self._find_token(approval_url)
                return PaypalPayment(payment.id, approval_url, token)
            else:
                raise PayPalError(payment.error)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def approve_payment(self, payment_id, payer_id):
        try:
            payment = paypal.Payment.find(payment_id)
            # When payment is not found PayPal returns empty result instead of raising an exception
            if not payment:
                raise PayPalError('Payment not found')
            if payment.execute({'payer_id': payer_id}):
                return True
            else:
                raise PayPalError(payment.error)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def create_plan(self, amount, tax, name, description, return_url, cancel_url):
        """
        Create and activate monthly billing plan.
        https://developer.paypal.com/docs/api/payments.billing-plans

        :param amount: Decimal value of plan payment for one month including tax.
        :param tax: Decimal value of VAT tax.
        :param name: Name of the billing plan.
        :param description: Description of the billing plan.
        :param return_url: Callback view URL for approved billing plan.
        :param cancel_url: Callback view URL for cancelled billing plan.
        :return: Billing plan ID.
        """
        if amount < tax:
            raise PayPalError('Plan price should be greater than tax.')

        plan = paypal.BillingPlan({
            'name': name,
            'description': description,
            'type': 'INFINITE',
            'payment_definitions': [{
                'name': 'Monthly payment for {}'.format(name),
                'type': 'REGULAR',
                'frequency_interval': 1,
                'frequency': 'MONTH',
                'cycles': 0,
                'amount': {
                    'currency': self.currency_name,
                    'value': self._format_decimal(amount - tax)
                },
                'charge_models': [
                    {
                        'type': 'TAX',
                        'amount': {
                            'currency': self.currency_name,
                            'value': self._format_decimal(tax)
                        }
                    }
                ]
            }],
            'merchant_preferences': {
                'return_url': return_url,
                'cancel_url': cancel_url,
                'auto_bill_amount': 'YES',
            }
        })

        try:
            if plan.create() and plan.activate():
                return plan.id
            else:
                raise PayPalError(plan.error)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def _format_decimal(self, value):
        """
        PayPal API expects at most two decimal places with a period separator.
        """
        return "%.2f" % value

    def _format_date(self, date):
        """
        At the moment timezone is ignored as only days of resources usage are counted, not hours.
        """
        return date.strftime('%Y-%m-%d UTC')

    def create_agreement(self, plan_id, name):
        """
        Create billing agreement. On success returns approval_url and token
        """
        # PayPal does not support immediate start of agreement
        # That's why we need to increase start date by small amount of time
        start_date = datetime.datetime.utcnow() + datetime.timedelta(minutes=1)

        # PayPal does not fully support ISO 8601 format
        formatted_date = start_date.strftime('%Y-%m-%dT%H:%M:%SZ')

        agreement = paypal.BillingAgreement({
            'name': name,
            'description': 'Agreement for {}'.format(name),
            'start_date': formatted_date,
            'payer': {'payment_method': 'paypal'},
            'plan': {'id': plan_id}
        })
        try:
            if agreement.create():
                approval_url = self._find_approval_url(agreement.links)

                # PayPal does not return agreement ID until it is approved
                # That's why we need to extract token in order to identify it with agreement in DB
                token = self._find_token(approval_url)
                return approval_url, token
            else:
                raise PayPalError(agreement.error)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def execute_agreement(self, payment_token):
        """
        Agreement should be executed if user has approved it.
        On success returns agreement id
        """
        try:
            agreement = paypal.BillingAgreement.execute(payment_token)
            if not agreement:
                raise PayPalError('Can not execute agreement')
            return agreement.id
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def get_agreement(self, agreement_id):
        """
        Get agreement from PayPal by ID
        """
        try:
            agreement = paypal.BillingAgreement.find(agreement_id)
            # When agreement is not found PayPal returns empty result instead of raising an exception
            if not agreement:
                raise PayPalError('Agreement not found')
            return agreement
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def cancel_agreement(self, agreement_id):
        agreement = self.get_agreement(agreement_id)

        try:
            # Because user may cancel agreement via PayPal web UI
            # we need to distinguish it from cancel done via API
            if agreement.cancel({'note': 'Canceling the agreement by application'}):
                return True
            else:
                raise PayPalError(agreement.error)
        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def get_agreement_transactions(self, agreement_id, start_date, end_date=None):
        if not end_date:
            end_date = timezone.now()

        # If start and end date are the same PayPal raises exceptions
        # That's why we need to increase end_date by one day
        if end_date - start_date < datetime.timedelta(days=1):
            end_date += datetime.timedelta(days=1)

        formatted_start_date = start_date.strftime('%Y-%m-%d')
        formatted_end_date = end_date.strftime('%Y-%m-%d')

        agreement = self.get_agreement(agreement_id)
        try:
            data = agreement.search_transactions(formatted_start_date, formatted_end_date)
            txs = data.agreement_transaction_list
            if not txs:
                return []

            results = []
            for tx in txs:
                if tx.status != 'Completed':
                    continue
                results.append({
                    'time_stamp': dateutil.parser.parse(tx.time_stamp),
                    'transaction_id': tx.transaction_id,
                    'amount': decimal.Decimal(tx.amount.value),
                    'payer_email': tx.payer_email
                })
            return results

        except paypal.exceptions.ConnectionError as e:
            raise PayPalError(e)

    def download_invoice_pdf(self, invoice):
        if not invoice.backend_id:
            raise PayPalError('Invoice for date %s and customer %s could not be found' % (
                invoice.invoice_date.strftime('%Y-%m-%d'),
                invoice.customer.name,
            ))

        invoice_url = self.get_payment_view_url(invoice.backend_id, {'printPdfMode': 'true'})
        response = urlopen(invoice_url)  # nosec
        content = response.read()
        invoice.pdf.save(invoice.file_name, ContentFile(content), save=True)



import os

from celery import Celery
from celery import signals

from waldur_core.logging.middleware import get_event_context, set_event_context, reset_event_context

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'waldur_core.server.settings')  # XXX:

app = Celery('waldur_core', namespace='CELERY', strict_typing=False)

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings')
app.autodiscover_tasks()


class PriorityRouter:
    """ Run heavy tasks and background tasks in separate queues. """

    def route_for_task(self, task_name, *args, **kwargs):
        task = app.tasks.get(task_name)
        if getattr(task, 'is_heavy_task', False):
            return {'queue': 'heavy'}
        if getattr(task, 'is_background', False):
            return {'queue': 'background'}
        return None


# The workflow for passing event context to background tasks works as following:
# 1) Generate event context at CaptureEventContextMiddleware and bind it to local thread
# 2) At the Django side: fetch event context from local thread and pass it as parameter to task
# 3) At Celery worker side: fetch event context from task and bind it to local thread
@signals.before_task_publish.connect
def pass_event_context(sender=None, body=None, **kwargs):
    if body is None:
        return

    event_context = get_event_context()
    if event_context:
        # kwargs is the second item in body tuple with index equal 1.
        # See also http://docs.celeryproject.org/en/v4.1.0/internals/protocol.html#version-2
        body[1]['event_context'] = event_context


@signals.task_prerun.connect
def bind_event_context(sender=None, **kwargs):
    try:
        event_context = kwargs['kwargs'].pop('event_context')
    except KeyError:
        return

    set_event_context(event_context)


@signals.task_postrun.connect
def unbind_event_context(sender=None, **kwargs):
    reset_event_context()



import logging

from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_mastermind.invoices import registrators
from waldur_mastermind.marketplace import models
from waldur_mastermind.marketplace_vmware import CPU_TYPE, RAM_TYPE, DISK_TYPE

logger = logging.getLogger(__name__)


def add_new_vm_to_invoice(sender, vm, **kwargs):
    registrators.RegistrationManager.register(vm, timezone.now())


def terminate_invoice_when_vm_deleted(sender, instance, **kwargs):
    registrators.RegistrationManager.terminate(instance, timezone.now())


def create_invoice_item_when_vm_is_updated(sender, vm, **kwargs):
    registrators.RegistrationManager.terminate(vm, timezone.now())
    registrators.RegistrationManager.register(vm, timezone.now())


def update_marketplace_resource_limits_when_vm_is_updated(sender, vm, **kwargs):
    try:
        resource = models.Resource.objects.get(scope=vm)
    except ObjectDoesNotExist:
        logger.debug('Skipping marketplace resource update for vm '
                     'because marketplace resource does not exist. '
                     'Resource ID: %s', core_utils.serialize_instance(vm))
    else:
        resource.limits[CPU_TYPE] = vm.cores
        resource.limits[RAM_TYPE] = vm.ram
        resource.limits[DISK_TYPE] = vm.total_disk
        resource.save()



class RancherException(Exception):
    pass


class BadRequest(RancherException):
    """General purpose exception class."""


class Unauthorized(BadRequest):
    """Raised when invalid credentials are provided."""

    message = 'Unauthorized: bad credentials.'



import datetime
import logging
import tarfile
import traceback

from celery import shared_task
from django.conf import settings
from django.contrib.contenttypes.models import ContentType

from waldur_core.core.utils import deserialize_instance
from waldur_core.logging.models import BaseHook, SystemNotification, Report, Feed, Event
from waldur_core.logging.utils import create_report_archive
from waldur_core.structure import models as structure_models

logger = logging.getLogger(__name__)


@shared_task(name='waldur_core.logging.process_event')
def process_event(event_id):
    event = Event.objects.get(id=event_id)
    for hook in BaseHook.get_active_hooks():
        if check_event(event, hook):
            hook.process(event)

    process_system_notification(event)


def process_system_notification(event):
    project_ct = ContentType.objects.get_for_model(structure_models.Project)
    project_feed = Feed.objects.filter(event=event, content_type=project_ct).first()
    project = project_feed and project_feed.scope

    customer_ct = ContentType.objects.get_for_model(structure_models.Customer)
    customer_feed = Feed.objects.filter(event=event, content_type=customer_ct).first()
    customer = customer_feed and customer_feed.scope

    for hook in SystemNotification.get_hooks(event.event_type, project=project, customer=customer):
        if check_event(event, hook):
            hook.process(event)


def check_event(event, hook):
    # Check that event matches with hook
    if event.event_type not in hook.all_event_types:
        return False

    # Check permissions
    for feed in Feed.objects.filter(event=event):
        qs = feed.content_type.model_class().get_permitted_objects(hook.user)
        if qs.filter(id=feed.object_id).exists():
            return True

    return False


@shared_task(name='waldur_core.logging.create_report')
def create_report(serialized_report):
    report = deserialize_instance(serialized_report)

    today = datetime.datetime.today()
    timestamp = today.strftime('%Y%m%dT%H%M%S')
    archive_filename = 'waldur-logs-%s-%s.tar.gz' % (timestamp, report.uuid.hex)

    try:
        cf = create_report_archive(
            settings.WALDUR_CORE['LOGGING_REPORT_DIRECTORY'],
            settings.WALDUR_CORE['LOGGING_REPORT_INTERVAL'],
        )
    except (tarfile.TarError, OSError, ValueError) as e:
        report.state = Report.States.ERRED
        error_message = 'Error message: %s. Traceback: %s' % (str(e), traceback.format_exc())
        report.error_message = error_message
        report.save()
    else:
        report.file.save(archive_filename, cf)
        report.file_size = cf.size
        report.state = Report.States.DONE
        report.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-28 20:40
from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0001_squashed_0054'),
        ('waldur_jira', '0006_add_parent_for_subtask'),
    ]

    operations = [
        migrations.CreateModel(
            name='Priority',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('icon_url', models.URLField(blank=True, verbose_name='icon url')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Priority',
                'verbose_name_plural': 'Priorities',
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.AlterField(
            model_name='issue',
            name='priority',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_jira.Priority'),
        ),
    ]



import calendar
from collections import OrderedDict
import datetime
import functools
import importlib
from itertools import chain
from operator import itemgetter
import os
import re
import time
import unicodedata
import uuid
import warnings

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.mail import EmailMultiAlternatives
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import F
from django.db.models.sql.query import get_order_dir
from django.http import QueryDict
from django.template import Context
from django.template.loader import get_template, render_to_string
from django.urls import resolve
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.encoding import force_text
import jwt
import requests
from requests.packages.urllib3 import exceptions
from rest_framework.settings import api_settings


def flatten(*xs):
    return tuple(chain.from_iterable(xs))


def sort_dict(unsorted_dict):
    """
    Return a OrderedDict ordered by key names from the :unsorted_dict:
    """
    sorted_dict = OrderedDict()
    # sort items before inserting them into a dict
    for key, value in sorted(unsorted_dict.items(), key=itemgetter(0)):
        sorted_dict[key] = value
    return sorted_dict


def format_time_and_value_to_segment_list(time_and_value_list, segments_count, start_timestamp,
                                          end_timestamp, average=False):
    """
    Format time_and_value_list to time segments

    Parameters
    ^^^^^^^^^^
    time_and_value_list: list of tuples
        Have to be sorted by time
        Example: [(time, value), (time, value) ...]
    segments_count: integer
        How many segments will be in result
    Returns
    ^^^^^^^
    List of dictionaries
        Example:
        [{'from': time1, 'to': time2, 'value': sum_of_values_from_time1_to_time2}, ...]
    """
    segment_list = []
    time_step = (end_timestamp - start_timestamp) / segments_count
    for i in range(segments_count):
        segment_start_timestamp = start_timestamp + time_step * i
        segment_end_timestamp = segment_start_timestamp + time_step
        value_list = [
            value for time, value in time_and_value_list
            if time >= segment_start_timestamp and time < segment_end_timestamp]
        segment_value = sum(value_list)
        if average and len(value_list) != 0:
            segment_value /= len(value_list)

        segment_list.append({
            'from': segment_start_timestamp,
            'to': segment_end_timestamp,
            'value': segment_value,
        })
    return segment_list


def datetime_to_timestamp(datetime):
    return int(time.mktime(datetime.timetuple()))


def timestamp_to_datetime(timestamp, replace_tz=True):
    dt = datetime.datetime.fromtimestamp(int(timestamp))
    if replace_tz:
        dt = dt.replace(tzinfo=timezone.get_current_timezone())
    return dt


def timeshift(**kwargs):
    return timezone.now().replace(microsecond=0) + datetime.timedelta(**kwargs)


def hours_in_month(month=None, year=None):
    now = datetime.datetime.now()
    if not month:
        month = now.month
    if not year:
        year = now.year

    days_in_month = calendar.monthrange(year, month)[1]
    return 24 * days_in_month


def month_start(date):
    return timezone.make_aware(datetime.datetime(day=1, month=date.month, year=date.year))


def month_end(date):
    days_in_month = calendar.monthrange(date.year, date.month)[1]
    last_day_of_month = datetime.date(month=date.month, year=date.year, day=days_in_month)
    last_second_of_month = datetime.datetime.combine(last_day_of_month, datetime.time.max)
    return timezone.make_aware(last_second_of_month, timezone.get_current_timezone())


def pwgen(pw_len=16):
    """ Generate a random password with the given length.
        Allowed chars does not have "I" or "O" or letters and
        digits that look similar -- just to avoid confusion.
    """
    return get_random_string(pw_len, 'abcdefghjkmnpqrstuvwxyz'
                                     'ABCDEFGHJKLMNPQRSTUVWXYZ'
                                     '23456789')


def serialize_instance(instance):
    """ Serialize Django model instance """
    model_name = force_text(instance._meta)
    return '{}:{}'.format(model_name, instance.pk)


def deserialize_instance(serialized_instance):
    """ Deserialize Django model instance """
    model_name, pk = serialized_instance.split(':')
    model = apps.get_model(model_name)
    return model._default_manager.get(pk=pk)


def serialize_class(cls):
    """ Serialize Python class """
    return '{}:{}'.format(cls.__module__, cls.__name__)


def deserialize_class(serilalized_cls):
    """ Deserialize Python class """
    module_name, cls_name = serilalized_cls.split(':')
    module = importlib.import_module(module_name)
    return getattr(module, cls_name)


def clear_url(url):
    """ Remove domain and protocol from url """
    if url.startswith('http'):
        return '/' + url.split('/', 3)[-1]
    return url


def get_model_from_resolve_match(match):
    queryset = match.func.cls.queryset
    if queryset is not None:
        return queryset.model
    else:
        return match.func.cls.model


def instance_from_url(url, user=None):
    """ Restore instance from URL """
    # XXX: This circular dependency will be removed then filter_queryset_for_user
    # will be moved to model manager method
    from waldur_core.structure.managers import filter_queryset_for_user

    url = clear_url(url)
    match = resolve(url)
    model = get_model_from_resolve_match(match)
    queryset = model.objects.all()
    if user is not None:
        queryset = filter_queryset_for_user(model.objects.all(), user)
    return queryset.get(**match.kwargs)


def get_detail_view_name(model):
    if model is NotImplemented:
        raise AttributeError('Cannot get detail view name for not implemented model')

    if hasattr(model, 'get_url_name') and callable(model.get_url_name):
        return '%s-detail' % model.get_url_name()

    return '%s-detail' % model.__name__.lower()


def get_list_view_name(model):
    if model is NotImplemented:
        raise AttributeError('Cannot get list view name for not implemented model')

    if hasattr(model, 'get_url_name') and callable(model.get_url_name):
        return '%s-list' % model.get_url_name()

    return '%s-list' % model.__name__.lower()


def get_fake_context():
    user = get_user_model()()
    request = type('R', (object,), {'method': 'GET', 'user': user, 'query_params': QueryDict()})
    return {'request': request, 'user': user}


def camel_case_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def silent_call(name, *args, **options):
    call_command(name, stdout=open(os.devnull, 'w'), *args, **options)


def format_text(template_name, context):
    template = get_template(template_name).template
    return template.render(Context(context, autoescape=False)).strip()


def send_mail_with_attachment(subject, body, to, from_email=None, html_message=None,
                              filename=None, attachment=None, content_type='text/plain'):
    from_email = from_email or settings.DEFAULT_FROM_EMAIL
    email = EmailMultiAlternatives(
        subject=subject,
        body=body,
        to=to,
        from_email=from_email
    )

    if html_message:
        email.attach_alternative(html_message, 'text/html')

    if filename:
        email.attach(filename, attachment, content_type)
    return email.send()


def broadcast_mail(app, event_type, context, recipient_list,
                   filename=None, attachment=None, content_type='text/plain'):
    """
    Shorthand to format email message from template file and sent it to all recipients.

    It is assumed that there are there are 3 templates available for event type in application.
    For example, if app is 'users' and event_type is 'invitation_rejected', then there should be 3 files:

    1) users/invitation_rejected_subject.txt is template for email subject
    2) users/invitation_rejected_message.txt is template for email body as text
    3) users/invitation_rejected_message.html is template for email body as HTML

    By default, built-in Django send_mail is used, all members
    of the recipient list will see the other recipients in the 'To' field.
    Contrary to this, we're using explicit loop in order to ensure that
    recipients would NOT see the other recipients.

    :param app: prefix for template filename.
    :param event_type: postfix for template filename.
    :param context: dictionary passed to the template for rendering.
    :param recipient_list: list of strings, each an email address.
    :param filename: name of the attached file
    :param attachment: content of attachment
    :param content_type: the content type of attachment
    """
    subject_template_name = '%s/%s_subject.txt' % (app, event_type)
    subject = format_text(subject_template_name, context)

    text_template_name = '%s/%s_message.txt' % (app, event_type)
    text_message = format_text(text_template_name, context)

    html_template_name = '%s/%s_message.html' % (app, event_type)
    html_message = render_to_string(html_template_name, context)

    for recipient in recipient_list:
        send_mail_with_attachment(subject, text_message, to=[recipient], html_message=html_message,
                                  filename=filename, attachment=attachment, content_type=content_type)


def get_ordering(request):
    """
    Extract ordering from HTTP request.
    """
    return request.query_params.get(api_settings.ORDERING_PARAM)


def order_with_nulls(queryset, field):
    """
    If sorting order is ascending, then NULL values come first,
    if sorting order is descending, then NULL values come last.
    """
    col, order = get_order_dir(field)
    descending = True if order == 'DESC' else False

    if descending:
        return queryset.order_by(F(col).desc(nulls_last=True))
    else:
        return queryset.order_by(F(col).asc(nulls_first=True))


def is_uuid_like(val):
    """
    Check if value looks like a valid UUID.
    """
    try:
        uuid.UUID(val)
    except (TypeError, ValueError, AttributeError):
        return False
    else:
        return True


def chunks(xs, n):
    """
    Split list to evenly sized chunks

    >> chunks(range(10), 4)
    [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9]]

    :param xs: arbitrary list
    :param n: chunk size
    :return: list of lists
    """
    return [xs[i:i + n] for i in range(0, len(xs), n)]


def create_batch_fetcher(fetcher):
    """
    Decorator to simplify code for chunked fetching.
    It fetches resources from backend API in evenly sized chunks.
    It is needed in order to avoid too long HTTP request error.
    Essentially, it gives the same result as fetcher(items) but does not throw an error.

    :param fetcher: fetcher: function which accepts list of items and returns list of results,
    for example, list of UUIDs and returns list of projects with given UUIDs
    :return: function with the same signature as fetcher
    """
    @functools.wraps(fetcher)
    def wrapped(items):
        """
        :param items: list of items for request, for example, list of UUIDs
        :return: merged list of results
        """
        result = []
        for chunk in chunks(items, settings.WALDUR_CORE['HTTP_CHUNK_SIZE']):
            result.extend(fetcher(chunk))
        return result
    return wrapped


class DryRunCommand(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('--dry-run', action='store_true',
                            help='Don\'t make any changes, instead show what objects would be created.')


def encode_jwt_token(data, api_secret_code=None):
    """
    Encode Python dictionary as JWT token.
    :param data: Dictionary with payload.
    :param api_secret_code: optional string, application secret key is used by default.
    :return: JWT token string with encoded and signed data.
    """
    if api_secret_code is None:
        api_secret_code = settings.SECRET_KEY
    return jwt.encode(data, api_secret_code, algorithm='HS256', json_encoder=DjangoJSONEncoder)


def decode_jwt_token(encoded_data, api_secret_code=None):
    """
    Decode JWT token string to Python dictionary.
    :param encoded_data: JWT token string with encoded and signed data.
    :param api_secret_code: optional string, application secret key is used by default.
    :return: Dictionary with payload.
    """
    if api_secret_code is None:
        api_secret_code = settings.SECRET_KEY
    return jwt.decode(encoded_data, api_secret_code, algorithms=['HS256'])


def normalize_unicode(data):
    return unicodedata.normalize(u'NFKD', data).encode('ascii', 'ignore').decode('utf8')


UNIT_PATTERN = re.compile('(\d+)([KMGTP]?)')

UNITS = {
    'K': 2**10,
    'M': 2**20,
    'G': 2**30,
    'T': 2**40,
}


def parse_int(value):
    """
    Convert 5K to 5000.
    """
    match = re.match(UNIT_PATTERN, value)
    if not match:
        return 0
    value = int(match.group(1))
    unit = match.group(2)
    if unit:
        factor = UNITS[unit]
    else:
        factor = 1
    return factor * value


class QuietSession(requests.Session):
    """Session class that suppresses warning about unsafe TLS sessions and clogging the logs.
    Inspired by: https://github.com/kennethreitz/requests/issues/2214#issuecomment-110366218
    """
    def request(self, *args, **kwargs):
        if not kwargs.get('verify', self.verify):
            with warnings.catch_warnings():
                if hasattr(exceptions, 'InsecurePlatformWarning'):  # urllib3 1.10 and lower does not have this warning
                    warnings.simplefilter('ignore', exceptions.InsecurePlatformWarning)
                warnings.simplefilter('ignore', exceptions.InsecureRequestWarning)
                return super(QuietSession, self).request(*args, **kwargs)
        else:
            return super(QuietSession, self).request(*args, **kwargs)



# Generated by Django 2.2.9 on 2020-02-07 11:21

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import waldur_rancher.models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0010_customer_geolocation'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('waldur_rancher', '0009_runtime_state'),
    ]

    operations = [
        migrations.CreateModel(
            name='RancherUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('is_active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='RancherUserClusterLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('role', waldur_rancher.models.ClusterRole(choices=[('owner', 'Cluster owner'), ('owner', 'Cluster member')], db_index=True, max_length=30)),
                ('cluster', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_rancher.Cluster')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_rancher.RancherUser')),
            ],
            options={
                'unique_together': {('user', 'cluster', 'role')},
            },
        ),
        migrations.AddField(
            model_name='rancheruser',
            name='clusters',
            field=models.ManyToManyField(through='waldur_rancher.RancherUserClusterLink', to='waldur_rancher.Cluster'),
        ),
        migrations.AddField(
            model_name='rancheruser',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='rancheruser',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name='rancheruser',
            unique_together={('user', 'settings')},
        ),
    ]



import re

from django.conf import settings
from django.core import exceptions, validators
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from waldur_core.core import models as core_models


def validate_username(value):
    if value in settings.WALDUR_FREEIPA['BLACKLISTED_USERNAMES']:
        raise exceptions.ValidationError(
            _('%(value)s is not valid FreeIPA username.'),
            params={'value': value},
        )


class Profile(core_models.UuidMixin, models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    username = models.CharField(
        _('username'), max_length=32, unique=True,
        help_text=_('Letters, numbers and ./+/-/_ characters'),
        validators=[
            validate_username,
            validators.RegexValidator(re.compile(r'^[a-zA-Z0-9_.][a-zA-Z0-9_.-]*[a-zA-Z0-9_.$-]?$'),
                                      _('Enter a valid username.'), 'invalid')
        ])
    agreement_date = models.DateTimeField(_('agreement date'), default=timezone.now,
                                          help_text=_('Indicates when the user has agreed with the policy.'))
    is_active = models.BooleanField(_('active'), default=True)
    tracker = FieldTracker()

    @property
    def gecos(self):
        param = []
        for field in ['full_name', 'email', 'phone_number']:
            value = getattr(self.user, field, None)
            if value:
                param.append(value)

        return ','.join(param)

    def __str__(self):
        return self.username



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-07-28 15:27
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('playbook_jobs', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='playbookparameter',
            options={'ordering': ['order']},
        ),
        migrations.AddField(
            model_name='playbookparameter',
            name='order',
            field=models.PositiveIntegerField(default=0),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-04 10:28
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0006_customer_backend_id'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='blocked',
            field=models.BooleanField(default=False),
        ),
    ]



import logging

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.lru_cache import lru_cache
from model_utils.models import TimeStampedModel

from waldur_core.core import models as core_models
from waldur_core.structure import SupportedServices, models as structure_models


logger = logging.getLogger(__name__)


def get_content_types_query(items):
    content_types = ContentType.objects.get_for_models(*items).values()
    return {'id__in': [ct.id for ct in content_types]}


@lru_cache(maxsize=1)
def get_service_content_types():
    services = [service['service'] for service in SupportedServices.get_service_models().values()]
    return get_content_types_query(services)


class DeploymentPlan(core_models.UuidMixin, core_models.NameMixin, TimeStampedModel):
    """
    Deployment plan contains list of plan items.
    """
    class Permissions:
        customer_path = 'project__customer'
        project_path = 'project'

    class Meta:
        ordering = ['-created']

    project = models.ForeignKey(on_delete=models.CASCADE, to=structure_models.Project, related_name='+')
    certifications = models.ManyToManyField(structure_models.ServiceCertification, blank=True)

    def __str__(self):
        return self.name

    @classmethod
    def get_url_name(cls):
        return 'deployment-plan'

    def get_requirements(self):
        """ Return how many ram, cores and storage are required for plan """
        requirements = {
            'ram': 0,
            'cores': 0,
            'storage': 0,
        }
        for item in self.items.all():
            requirements['ram'] += item.preset.ram * item.quantity
            requirements['cores'] += item.preset.cores * item.quantity
            requirements['storage'] += item.preset.storage * item.quantity
        return requirements

    def get_required_certifications(self):
        return set(list(self.certifications.all()) + list(self.project.certifications.all()))


class DeploymentPlanItem(models.Model):
    """
    Plan item specifies quantity of presets.

    For example:
    {
        "preset": <Hadoop DataNode>,
        "quantity": 10
    }
    """
    class Meta:
        ordering = 'plan', 'preset'
        unique_together = 'plan', 'preset'

    plan = models.ForeignKey(on_delete=models.CASCADE, to=DeploymentPlan, related_name='items')
    preset = models.ForeignKey(on_delete=models.CASCADE, to='Preset')
    quantity = models.PositiveSmallIntegerField(default=1)

    def __str__(self):
        return '%s %s' % (self.quantity, self.preset)


class Category(core_models.NameMixin):
    class Meta:
        verbose_name_plural = 'Categories'

    def __str__(self):
        return self.name


class Preset(core_models.UuidMixin, core_models.NameMixin):
    """
    Resource configuration preset.

    Example rendering of preset:
    {
        "category": "Big Data",
        "name": "Hadoop DataNode",
        "variant": "Large",
        "ram": "10240",
        "cores": "16",
        "storage": "1024000",
    }
    """
    class Meta:
        ordering = 'category', 'name', 'variant'
        unique_together = 'category', 'name', 'variant'

    SMALL = 'small'
    MEDIUM = 'medium'
    LARGE = 'large'

    VARIANTS = (
        (SMALL, 'Small'),
        (MEDIUM, 'Medium'),
        (LARGE, 'Large'),
    )

    category = models.ForeignKey(on_delete=models.CASCADE, to=Category, related_name='presets')
    variant = models.CharField(max_length=150, choices=VARIANTS)
    ram = models.PositiveIntegerField(default=0)
    cores = models.PositiveIntegerField(default=0, help_text='Preset cores count.')
    storage = models.PositiveIntegerField(default=0)

    def __str__(self):
        return '%s %s %s' % (self.variant, self.name, self.category)

    @classmethod
    def get_url_name(cls):
        return 'deployment-preset'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-07-03 09:12
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0009_project_is_removed'),
        ('waldur_vmware', '0008_network'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerNetwork',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer')),
                ('network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_vmware.Network')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='customernetwork',
            unique_together=set([('customer', 'network')]),
        ),
    ]



from django_filters.rest_framework import DjangoFilterBackend

from waldur_core.structure import models as structure_models


class OfferingCustomersFilterBackend(DjangoFilterBackend):
    def filter_queryset(self, request, queryset, view):
        user = request.user
        if user.is_staff:
            return queryset
        else:
            customers = structure_models.CustomerPermission.objects.filter(
                user=user,
                role=structure_models.CustomerRole.OWNER).values_list('customer', flat=True)
            return queryset.filter(offering__customer_id__in=customers)



from django.db import transaction
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import permissions as structure_permissions, models as structure_models

from . import models, register


class PresetSerializer(serializers.HyperlinkedModelSerializer):
    category = serializers.ReadOnlyField(source='category.name')
    variant = serializers.ReadOnlyField(source='get_variant_display')

    class Meta:
        model = models.Preset
        fields = ('url', 'uuid', 'name', 'category', 'variant', 'ram', 'cores', 'storage')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'deployment-preset-detail'},
        }


class DeploymentPlanItemSerializer(serializers.ModelSerializer):
    preset = PresetSerializer()

    class Meta:
        model = models.DeploymentPlanItem
        fields = ('preset', 'quantity',)


class NestedDeploymentPlanItemSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.DeploymentPlanItem
        fields = ('preset', 'quantity')
        extra_kwargs = {
            'preset': {
                'lookup_field': 'uuid',
                'view_name': 'deployment-preset-detail'
            }
        }


class NestedCertificatesSerializer(core_serializers.HyperlinkedRelatedModelSerializer):
    class Meta:
        model = structure_models.ServiceCertification
        fields = ('url', 'uuid', 'name', 'description', 'link')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'service-certification-detail'},
        }


class BaseDeploymentPlanSerializer(core_serializers.AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):
    certifications = NestedCertificatesSerializer(
        many=True, queryset=structure_models.ServiceCertification.objects.all(), required=False)

    class Meta:
        model = models.DeploymentPlan
        fields = ('url', 'uuid', 'name', 'project', 'items', 'certifications')
        protected_fields = ('project',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'project': {'lookup_field': 'uuid'},
        }


class DeploymentPlanSerializer(BaseDeploymentPlanSerializer):
    items = DeploymentPlanItemSerializer(many=True)


class DeploymentPlanCreateSerializer(BaseDeploymentPlanSerializer):
    items = NestedDeploymentPlanItemSerializer(many=True, required=False)

    def validate_project(self, project):
        structure_permissions.is_administrator(self.context['request'], self.context['view'], project)
        return project

    def create(self, validated_data):
        items = validated_data.pop('items', [])
        certifications = validated_data.pop('certifications', [])
        plan = super(DeploymentPlanCreateSerializer, self).create(validated_data)
        for item in items:
            plan.items.create(**item)
        plan.certifications.add(*certifications)
        return plan

    def update(self, instance, validated_data):
        items = validated_data.pop('items', None)
        certifications = validated_data.pop('certifications', None)

        plan = super(DeploymentPlanCreateSerializer, self).update(instance, validated_data)

        if certifications is not None:
            with transaction.atomic():
                plan.certifications.clear()
                plan.certifications.add(*certifications)

        if items is None:
            return plan

        current_map = {item.preset_id: item.quantity for item in plan.items.all()}
        current_ids = set(current_map.keys())

        new_map = {item['preset'].id: item['quantity'] for item in items}
        new_ids = set(new_map.keys())

        with transaction.atomic():
            # Remove stale items
            plan.items.filter(preset_id__in=current_ids - new_ids).delete()

            # Create new items
            for item_id in new_ids - current_ids:
                plan.items.create(preset_id=item_id, quantity=new_map[item_id])

            # Update existing items
            for item_id in new_ids & current_ids:
                plan.items.filter(preset_id=item_id).update(quantity=new_map[item_id])

        return plan


class OptimizedServiceSummarySerializer(serializers.Serializer):
    """ Serializer that renders each instance with its own specific serializer """

    @classmethod
    def get_serializer(cls, optimized_service):
        if optimized_service.error_message:
            return OptimizedServiceSerializer
        return register.Register.get_serilizer(optimized_service.service.settings.type) or OptimizedServiceSerializer

    def to_representation(self, instance):
        serializer = self.get_serializer(instance)
        return serializer(instance, context=self.context).data


class OptimizedServiceSerializer(serializers.Serializer):
    price = serializers.DecimalField(max_digits=22, decimal_places=10)
    service_settings = serializers.HyperlinkedRelatedField(
        source='service.settings',
        view_name='servicesettings-detail',
        lookup_field='uuid',
        read_only=True,
    )
    service_settings_name = serializers.ReadOnlyField(source='service.settings.name')
    service_settings_type = serializers.ReadOnlyField(source='service.settings.type')
    error_message = serializers.ReadOnlyField()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-12 14:53
from django.db import migrations
from django.db.models import F


PACKAGE_TYPE = 'Packages.Template'
RAM_TYPE = 'ram'
CORES_TYPE = 'cores'
STORAGE_TYPE = 'storage'


def migrate_components(apps, schema_editor):
    PlanComponent = apps.get_model('marketplace', 'PlanComponent')
    queryset = PlanComponent.objects.filter(
        component__offering__type=PACKAGE_TYPE,
        component__type__in=(RAM_TYPE, STORAGE_TYPE)
    )
    # In order to avoid numeric overflow we need to ensure that there are no invalid values.
    # A field with precision 14, scale 10 must round to an absolute value less than 10^4.
    queryset.filter(price__gt=9).update(price=0)
    queryset.update(
        amount=F('amount') / 1024,
        price=F('price') * 1024
    )


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0001_squashed_0093'),
        ('packages', '0001_squashed_0015'),
    ]

    operations = [
        migrations.RunPython(migrate_components)
    ]



from django.apps import AppConfig
from django.db.models import signals


class RancherInvoicesConfig(AppConfig):
    name = 'waldur_mastermind.rancher_invoices'
    verbose_name = 'Rancher'

    def ready(self):
        from waldur_rancher import models as rancher_models, signals as rancher_signals
        from waldur_mastermind.marketplace import models as marketplace_models
        from . import handlers

        rancher_signals.node_states_have_been_updated.connect(
            handlers.update_node_usage,
            sender=rancher_models.Cluster,
            dispatch_uid='support_invoices.handlers.update_node_usage',
        )

        signals.post_save.connect(
            handlers.create_invoice_item_if_component_usage_has_been_created,
            sender=marketplace_models.ComponentUsage,
            dispatch_uid='support_invoices.handlers.create_invoice_item_if_component_usage_has_been_created',
        )



from . import views


def register_in(router):
    router.register(r'vmware', views.ServiceViewSet, basename='vmware')
    router.register(r'vmware-limits', views.LimitViewSet, basename='vmware-limit')
    router.register(r'vmware-service-project-link', views.ServiceProjectLinkViewSet,
                    basename='vmware-spl')
    router.register(r'vmware-virtual-machine', views.VirtualMachineViewSet,
                    basename='vmware-virtual-machine')
    router.register(r'vmware-disks', views.DiskViewSet,
                    basename='vmware-disk')
    router.register(r'vmware-ports', views.PortViewSet,
                    basename='vmware-port')
    router.register(r'vmware-templates', views.TemplateViewSet,
                    basename='vmware-template')
    router.register(r'vmware-clusters', views.ClusterViewSet,
                    basename='vmware-cluster')
    router.register(r'vmware-networks', views.NetworkViewSet,
                    basename='vmware-network')
    router.register(r'vmware-datastores', views.DatastoreViewSet,
                    basename='vmware-datastore')
    router.register(r'vmware-folders', views.FolderViewSet,
                    basename='vmware-folder')
